# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
# TODO: make the datacube configurable from outside
# TODO: make the datacube based on two other webservices: iso and wms

from pystac.extensions.datacube import (
    AdditionalDimension,
    DatacubeExtension,
    DimensionType,
    HorizontalSpatialDimension,
    TemporalDimension,
    Variable,
    VariableType,
    VerticalSpatialDimension,
)


class Datacube(object):
    def item_extension(self, item, harvesting_vars):
        variables = (
            {}
        )  # variables dictionary for gathering the Variable objects
        dimensions = (
            {}
        )  # dimensions dictionary for gathering the Dimension objects
        cube = DatacubeExtension.ext(item, add_if_missing=True)

        variable_dimensions = [
            elem.replace(" ", ",").split(",") if " " in elem else [elem]
            for elem in harvesting_vars["variable_dimensions"]
        ]
        if len(harvesting_vars["variable_ids"]) != len(variable_dimensions):
            # Another solution for this is to index each output element and find the None values to decide about them later
            raise ValueError(
                "The length of the variable list and the dimension list are not equal. Check your dataset in Thredds. WE SHOULD EXPLAIN THIS ERROR IN THE DOCUMENTATION"
            )
        else:
            for i, v in enumerate(harvesting_vars["variable_ids"]):
                variable_dict = dict()
                # forth case is when description is not available or is less than then variable ids or dimensions. It's not required then can be ignored
                # Usually every varialbe in ncML has name and shape attrs that it can be used as variable id and dimension.

                variable_dict["dimensions"] = variable_dimensions[i]
                variable_dict["type"] = VariableType.DATA.value
                if (
                    harvesting_vars["variable_description"] is not None
                    and len(harvesting_vars["variable_ids"])
                    == len(harvesting_vars["variable_description"])
                    and len(harvesting_vars["variable_description"])
                    != harvesting_vars["variable_description"].count(None)
                ):
                    if harvesting_vars["variable_description"][i] is not None:
                        variable_dict["description"] = harvesting_vars[
                            "variable_description"
                        ][i]
                variable_dict["dimensions"] = variable_dimensions[i]
                if (
                    harvesting_vars["variable_unit"] is not None
                    and len(harvesting_vars["variable_unit"])
                    == len(variable_dimensions)
                    and len(harvesting_vars["variable_unit"])
                    != harvesting_vars["variable_unit"].count(None)
                ):
                    if harvesting_vars["variable_unit"][i] is not None:
                        variable_dict["units"] = harvesting_vars[
                            "variable_unit"
                        ][i]
                variables[harvesting_vars["variable_ids"][i]] = Variable(
                    variable_dict
                )

        # Horizontal Spatial Dimension
        list_of_required_keys = [
            "horizontal_ids_lat",
            "horizontal_ids_lon",
            "horizontal_axis_x",
            "horizontal_axis_y",
            "horizontal_extent_lon_min",
            "horizontal_extent_lon_max",
            "horizontal_extent_lat_min",
            "horizontal_extent_lat_max",
        ]
        # TODO: check the possiblity of making the string values as float
        if all(
            harvesting_vars.get(key) is not None
            for key in list_of_required_keys
        ):
            print("test")
            horizontal_dict = dict()
            horizontal_dict = {
                "type": DimensionType.SPATIAL.value,
                "axis": harvesting_vars["horizontal_axis_x"],
                "description": harvesting_vars["horizontal_description_lon"],
                "extent": [
                    float(harvesting_vars["horizontal_extent_lon_min"]),
                    float(harvesting_vars["horizontal_extent_lon_max"]),
                ],
                "reference_system": harvesting_vars[
                    "horizontal_reference_system"
                ],
            }
            dimensions[
                harvesting_vars["horizontal_ids_lon"]
            ] = HorizontalSpatialDimension(horizontal_dict)
            horizontal_dict = dict()
            horizontal_dict = {
                "type": DimensionType.SPATIAL.value,
                "axis": harvesting_vars["horizontal_axis_y"],
                "description": harvesting_vars["horizontal_description_lat"],
                "extent": [
                    float(harvesting_vars["horizontal_extent_lat_min"]),
                    float(harvesting_vars["horizontal_extent_lat_max"]),
                ],
                "reference_system": harvesting_vars[
                    "horizontal_reference_system"
                ],
            }
            dimensions[
                harvesting_vars["horizontal_ids_lat"]
            ] = HorizontalSpatialDimension(horizontal_dict)
        else:
            print(
                "Required attributes are not involved in the output dictionary. Check your dataset in Thredds or config json file. WE SHOULD EXPLAIN THIS ERROR IN THE DOCUMENTATION"
            )

        # Vertical Spatial Dimension
        list_of_required_keys = [
            "vertical_id",
            "vertical_axis",
            "vertical_extent_upper",
            "vertical_extent_lower",
        ]
        if all(
            harvesting_vars.get(key) is not None and harvesting_vars[key] != []
            for key in list_of_required_keys
        ):
            vertical_dict = dict()
            vertical_dict = {
                "type": DimensionType.SPATIAL.value,
                "axis": harvesting_vars["vertical_axis"],
                "description": harvesting_vars["vertical_description"],
                "extent": [
                    float(harvesting_vars["vertical_extent_lower"]),
                    float(harvesting_vars["vertical_extent_upper"]),
                ],
            }
            dimensions[
                harvesting_vars["vertical_id"]
            ] = VerticalSpatialDimension(vertical_dict)
        else:
            print(
                "Required attributes are not involved in the output dictionary. Check your dataset in Thredds or config json file. WE SHOULD EXPLAIN THIS ERROR IN THE DOCUMENTATION"
            )

        # Temporal Dimension
        list_of_required_keys = [
            "temporal_id",
            "temporal_extent_start_datetime",
            "temporal_extent_end_datetime",
        ]
        if all(
            harvesting_vars.get(key) is not None
            for key in list_of_required_keys
        ):
            temporal_dict = dict()
            temporal_dict = {
                "type": DimensionType.TEMPORAL.value,
                "description": harvesting_vars["temporal_description"],
                "extent": [
                    harvesting_vars["temporal_extent_start_datetime"],
                    harvesting_vars["temporal_extent_end_datetime"],
                ],
            }
            dimensions[harvesting_vars["temporal_id"]] = TemporalDimension(
                temporal_dict
            )
        else:
            print(
                "Required attributes are not involved in the output dictionary. Check your dataset in Thredds or config json file. WE SHOULD EXPLAIN THIS ERROR IN THE DOCUMENTATION"
            )

        # Additional Dimensions
        anotherlist = [
            "x",
            "y",
            "time",
            "longitude",
            "latitude",
            "t",
            "z",
            "height",
        ]
        for id in harvesting_vars["additional_ids"]:
            if id.lower() not in [
                i.lower() for i in list(dimensions.keys()) + anotherlist
            ]:
                additional_dict = dict()
                additional_dict = {
                    "type": harvesting_vars["additional_type"],
                    "extent": [None, None],
                }
                dimensions[id] = AdditionalDimension(additional_dict)
        cube.apply(dimensions=dimensions, variables=variables)

    # def item(item, harvesting_vars):
    #     var = []
    #     dim = []
    #     variables = {}  # variables in a STAC-Item
    #     dimensions = {}  # dimensions in a STAC-Item
    #     cube = DatacubeExtension.ext(item, add_if_missing=True)
    #     # Creating dimension and varibles to datacube extension
    #     # TODO: temporary we add the following line to avoid error in STAC validation but we need to find a better solution
    #     # ISSUE #81
    #     # print('harvesting_vars["keyword"]', harvesting_vars["keyword"])
    #     # print('harvesting_vars["var_lists"]', harvesting_vars["var_lists"])
    #     # print('harvesting_vars["var_descs"]', harvesting_vars["var_descs"])
    #     # print('harvesting_vars["var_dims"]', harvesting_vars["var_dims"])
    #     DatacubeExtension.remove_from(item)
    #     for i, v in enumerate(harvesting_vars["var_lists"]):
    #         if harvesting_vars["keyword"] == []:
    #             var.append(
    #                 Variable(
    #                     dict(
    #                         type="data",
    #                         description=harvesting_vars["var_descs"][i],
    #                         dimensions=harvesting_vars["keyword"],
    #                     )
    #                 )
    #             )
    #         else:
    #             var.append(
    #                 Variable(
    #                     dict(
    #                         type="data",
    #                         description=harvesting_vars["var_descs"][i],
    #                         dimensions=harvesting_vars["keyword"][i],
    #                     )
    #                 )
    #             )

    #     for i, v in enumerate(harvesting_vars["var_lists"]):
    #         if harvesting_vars["var_lists"][i] not in harvesting_vars["var_dims"] and harvesting_vars["var_lists"][i] not in harvesting_vars["keyword"]:
    #             variables[v] = var[i]

    #     for i, v in enumerate(harvesting_vars["var_dims"]):
    #         if harvesting_vars["var_dims"][i] == "row" or "lon" in harvesting_vars["var_dims"][i]:
    #             extend_ = [
    #                 harvesting_vars["long_min"].replace(",", "."),
    #                 harvesting_vars["long_max"].replace(",", "."),
    #             ]
    #             type_ = DimensionType.SPATIAL.value
    #             description_ = "longitude"
    #             axis_ = "x"
    #         elif harvesting_vars["var_dims"][i] == "column" or "lat" in harvesting_vars["var_dims"][i]:
    #             extend_ = [
    #                 harvesting_vars["lat_min"].replace(",", "."),
    #                 harvesting_vars["lat_max"].replace(",", "."),
    #             ]
    #             type_ = DimensionType.SPATIAL.value
    #             description_ = "latitude"
    #             axis_ = "y"
    #         elif harvesting_vars["var_dims"][i] == "temporal" or "time" in harvesting_vars["var_dims"][i] or "t" in harvesting_vars["var_dims"][i]:
    #             extend_ = [
    #                 harvesting_vars["time_start"],
    #                 harvesting_vars["time_end"],
    #             ]
    #             type_ = DimensionType.TEMPORAL.value
    #             description_ = "time"
    #             axis_ = "time"
    #         else:
    #             extend_ = None
    #             type_ = DimensionType.SPATIAL.value
    #             description_ = harvesting_vars["var_dims"][i]
    #             axis_ = harvesting_vars["var_dims"][i]

    #         dim.append(
    #             HorizontalSpatialDimension(
    #                 properties=dict(
    #                     axis=axis_,
    #                     extent=extend_,
    #                     description=description_,
    #                     reference_system="epsg:4326",
    #                     type=type_,
    #                 )
    #             )
    #         )
    #     for i, v in enumerate(harvesting_vars["var_dims"]):
    #         dimensions[v] = dim[i]
    #     cube.apply(dimensions=dimensions, variables=variables)



import os
from datetime import datetime

# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
import pystac

from .. import utils

unidata = "http://www.unidata.ucar.edu/namespaces/thredds/InvCatalog/v1.0"
w3 = "http://www.w3.org/1999/xlink"
iso_gmd = "http://www.isotc211.org/2005/gmd"
iso_gco = "http://www.isotc211.org/2005/gco"
iso_gml = "http://www.opengis.net/gml/3.2"
global_bounding_box = [-360, -90, 0, 90]
No_inf = "No information"
ncml = "http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2"
wms = "http://www.opengis.net/wms"
# a list of parameters to be not used in 'aName' parameter in iso webservice and 'dimension' and 'variable' parameters in ncML webservice
avoided_list = ["time_bnds", "bnds", "ens"]
# a list of parameters to be used in 'keyword' parameter in iso webservice
allowed_list = [
    "time",
    "lat",
    "latitude",
    "lon",
    "longitude",
    "long",
    "time_bnds",
    "bnds",
]
# a list of parameters to avoid use of them in final item
avoided_formats = [
    "float",
    "double",
    "int",
    "time_bnds",
    "bnds",
    "ens",
    "String",
    "short",
]
# Deafult value for having a bounding box for time-series data.
epilon = 0.000001


# webservice types that TDS2STAC supports
webservices_list = ["ncml", "wms", "iso"]
# List of array value attributes in `auto_switch_harvester`
append_list = ["keyword", "var_lists", "var_descs", "var_dims"]
# List of single value attributes in `auto_switch_harvester`
constant_list = [
    "long_min",
    "long_max",
    "lat_min",
    "lat_max",
    "time_start",
    "time_end",
]
# In the loop of 'auto_switch_harvester', criteria for locating the correct tag.
webservices_constants = dict(
    iso={
        "long_min": "{%s}westBoundLongitude_long_min" % iso_gmd,
        "long_max": "{%s}eastBoundLongitude_long_max" % iso_gmd,
        "lat_min": "{%s}southBoundLatitude_lat_min" % iso_gmd,
        "lat_max": "{%s}northBoundLatitude_lat_max" % iso_gmd,
        "keyword": "{%s}dimensionName_keyword" % iso_gmd,
        "var_lists": "{%s}MemberName_var_lists" % iso_gco,
        "var_descs": "{%s}descriptor_var_descs" % iso_gmd,
        "var_dims": "{%s}dimensionName_var_dims" % iso_gmd,
        "time_start": "{%s}beginPosition_time_start" % iso_gml,
        "time_end": "{%s}endPosition_time_end" % iso_gml,
    },
    ncml={
        "long_min": "geospatial_lon_min_long_min",
        "long_max": "geospatial_lon_max_long_max",
        "lat_min": "geospatial_lat_min_lat_min",
        "lat_max": "geospatial_lat_max_lat_max",
        "keyword": "{%s}variable_keyword" % ncml,
        "var_lists": "{%s}variable_var_lists" % ncml,
        "var_descs": "{%s}variable_var_descs" % ncml,
        "var_dims": "{%s}dimension_var_dims" % ncml,
        "time_start": "time_coverage_start_time_start",
        "time_end": "time_coverage_end_time_end",
    },
    wms={
        "long_min": "{%s}westBoundLongitude_long_min" % wms,
        "long_max": "{%s}eastBoundLongitude_long_max" % wms,
        "lat_min": "{%s}southBoundLatitude_lat_min" % wms,
        "lat_max": "{%s}northBoundLatitude_lat_max" % wms,
        "keyword": "{%s}Layer_keyword" % wms,
        "var_lists": "{%s}Layer_var_lists" % wms,
        "var_descs": "{%s}Layer_var_descs" % wms,
        "var_dims": "{%s}Layer_var_dims" % wms,
        "time_start": "{%s}Dimension_time_start" % wms,
        "time_end": "{%s}Dimension_time_end" % wms,
    },
)

schemas_dicts = {
    "catalog": "http://www.unidata.ucar.edu/namespaces/thredds/InvCatalog/v1.0",
    "dap4": "http://xml.opendap.org/ns/DAP/4.0#",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    "gco": "http://www.isotc211.org/2005/gco",
    "gmd": "http://www.isotc211.org/2005/gmd",
    "gmi": "http://www.isotc211.org/2005/gmi",
    "srv": "http://www.isotc211.org/2005/srv",
    "gmx": "http://www.isotc211.org/2005/gmx",
    "gsr": "http://www.isotc211.org/2005/gsr",
    "gss": "http://www.isotc211.org/2005/gss",
    "gts": "http://www.isotc211.org/2005/gts",
    "gml": "http://www.opengis.net/gml/3.2",
    "xs": "http://www.w3.org/2001/XMLSchema",
    "ncml": "http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2",
    "wms": "http://www.opengis.net/wms",
    "xlink": "http://www.w3.org/1999/xlink",
    "edal": "http://reading-escience-centre.github.io/edal-java/wms",
}


static_list_webservices = [
    "tds2stac_mode_analyser",
    "tds2stac_manual_variable",
    "tds2stac_webservice_analyser",
    "tds2stac_reference_key",
]

default_tag_config_file = (
    os.path.dirname(os.path.abspath(utils.__file__)) + "/tag_config.json"
)
webservices_constants = dict(
    iso={
        "long_min": "{%s}westBoundLongitude_long_min" % iso_gmd,
        "long_max": "{%s}eastBoundLongitude_long_max" % iso_gmd,
        "lat_min": "{%s}southBoundLatitude_lat_min" % iso_gmd,
        "lat_max": "{%s}northBoundLatitude_lat_max" % iso_gmd,
        "keyword": "{%s}dimensionName_keyword" % iso_gmd,
        "var_lists": "{%s}MemberName_var_lists" % iso_gco,
        "var_descs": "{%s}descriptor_var_descs" % iso_gmd,
        "var_dims": "{%s}dimensionName_var_dims" % iso_gmd,
        "time_start": "{%s}beginPosition_time_start" % iso_gml,
        "time_end": "{%s}endPosition_time_end" % iso_gml,
    },
    ncml={
        "long_min": "geospatial_lon_min_long_min",
        "long_max": "geospatial_lon_max_long_max",
        "lat_min": "geospatial_lat_min_lat_min",
        "lat_max": "geospatial_lat_max_lat_max",
        "keyword": "{%s}variable_keyword" % ncml,
        "var_lists": "{%s}variable_var_lists" % ncml,
        "var_descs": "{%s}variable_var_descs" % ncml,
        "var_dims": "{%s}dimension_var_dims" % ncml,
        "time_start": "time_coverage_start_time_start",
        "time_end": "time_coverage_end_time_end",
    },
    wms={
        "long_min": "{%s}westBoundLongitude_long_min" % wms,
        "long_max": "{%s}eastBoundLongitude_long_max" % wms,
        "lat_min": "{%s}southBoundLatitude_lat_min" % wms,
        "lat_max": "{%s}northBoundLatitude_lat_max" % wms,
        "keyword": "{%s}Layer_keyword" % wms,
        "var_lists": "{%s}Layer_var_lists" % wms,
        "var_descs": "{%s}Layer_var_descs" % wms,
        "var_dims": "{%s}Layer_var_dims" % wms,
        "time_start": "{%s}Dimension_time_start" % wms,
        "time_end": "{%s}Dimension_time_end" % wms,
    },
)


colorbars = [
    "alg",
    "alg-inv",
    "alg2",
    "alg2-inv",
    "bluered_colorbrewer",
    "bluered_colorbrewer-inv",
    "cdi",
    "cdi-inv",
    "default",
    "default-inv",
    "div-BrBG",
    "div-BrBG-inv",
    "div-BuRd",
    "div-BuRd-inv",
    "div-BuRd2",
    "div-BuRd2-inv",
    "div-PRGn",
    "div-PRGn-inv",
    "div-PiYG",
    "div-PiYG-inv",
    "div-PuOr",
    "div-PuOr-inv",
    "div-RdBu",
    "div-RdBu-inv",
    "div-RdGy",
    "div-RdGy-inv",
    "div-RdYlBu",
    "div-RdYlBu-inv",
    "div-RdYlGn",
    "div-RdYlGn-inv",
    "div-Spectral",
    "div-Spectral-inv",
    "ferret",
    "ferret-inv",
    "greyscale",
    "greyscale-inv",
    "lai",
    "lai-inv",
    "ncview",
    "ncview-inv",
    "ndvi",
    "ndvi-inv",
    "npp",
    "npp-inv",
    "npp_sv",
    "npp_sv-inv",
    "occam",
    "occam-inv",
    "occam_pastel-30",
    "occam_pastel-30-inv",
    "precip",
    "precip-inv",
    "psu-inferno",
    "psu-inferno-inv",
    "psu-magma",
    "psu-magma-inv",
    "psu-plasma",
    "psu-plasma-inv",
    "psu-viridis",
    "psu-viridis-inv",
    "redblue",
    "redblue-inv",
    "redblue_colorbrewer",
    "redblue_colorbrewer-inv",
    "seq-BkBu",
    "seq-BkBu-inv",
    "seq-BkGn",
    "seq-BkGn-inv",
    "seq-BkRd",
    "seq-BkRd-inv",
    "seq-BkYl",
    "seq-BkYl-inv",
    "seq-BlueHeat",
    "seq-BlueHeat-inv",
    "seq-Blues",
    "seq-Blues-inv",
    "seq-BuGn",
    "seq-BuGn-inv",
    "seq-BuPu",
    "seq-BuPu-inv",
    "seq-BuYl",
    "seq-BuYl-inv",
    "seq-GnBu",
    "seq-GnBu-inv",
    "seq-Greens",
    "seq-Greens-inv",
    "seq-Greys",
    "seq-Greys-inv",
    "seq-GreysRev",
    "seq-GreysRev-inv",
    "seq-Heat",
    "seq-Heat-inv",
    "seq-OrRd",
    "seq-OrRd-inv",
    "seq-Oranges",
    "seq-Oranges-inv",
    "seq-PuBu",
    "seq-PuBu-inv",
    "seq-PuBuGn",
    "seq-PuBuGn-inv",
    "seq-PuRd",
    "seq-PuRd-inv",
    "seq-Purples",
    "seq-Purples-inv",
    "seq-RdPu",
    "seq-RdPu-inv",
    "seq-Reds",
    "seq-Reds-inv",
    "seq-YlGn",
    "seq-YlGn-inv",
    "seq-YlGnBu",
    "seq-YlGnBu-inv",
    "seq-YlOrBr",
    "seq-YlOrBr-inv",
    "seq-YlOrRd",
    "seq-YlOrRd-inv",
    "seq-cubeYF",
    "seq-cubeYF-inv",
    "sld",
    "sld-inv",
    "sld2",
    "sld2-inv",
    "sm_categ_extreme",
    "sm_categ_extreme-inv",
    "sm_categ_quintile",
    "sm_categ_quintile-inv",
    "sm_categ_tercile",
    "sm_categ_tercile-inv",
    "smi",
    "smi-inv",
    "spi",
    "spi-inv",
    "spi_alternative",
    "spi_alternative-inv",
    "sst_36",
    "sst_36-inv",
    "temp_categ_extreme",
    "temp_categ_extreme-inv",
    "temp_categ_quintile",
    "temp_categ_quintile-inv",
    "temp_categ_tercile",
    "temp_categ_tercile-inv",
    "test3",
    "test3-inv",
    "tp_above_probab",
    "tp_above_probab-inv",
    "tp_below_probab",
    "tp_below_probab-inv",
    "tp_categ_extreme",
    "tp_categ_extreme-inv",
    "tp_categ_quintile",
    "tp_categ_quintile-inv",
    "tp_categ_tercile",
    "tp_categ_tercile-inv",
    "tp_extreme_dry_probab",
    "tp_extreme_dry_probab-inv",
    "tp_extreme_wet_probab",
    "tp_extreme_wet_probab-inv",
    "whiteblue10",
    "whiteblue10-inv",
    "whitered10",
    "whitered10-inv",
    "wms_precip",
    "wms_precip-inv",
    "x-Ncview",
    "x-Ncview-inv",
    "x-Occam",
    "x-Occam-inv",
    "x-Rainbow",
    "x-Rainbow-inv",
    "x-Sst",
    "x-Sst-inv",
]

# harvesting_vars = {  # type: ignore
#     "item_id": None,
#     "long_min": None,
#     "long_max": None,
#     "lat_min": None,
#     "lat_max": None,
#     "keyword": [],  # dimention of each variable
#     "var_lists": [],
#     "var_descs": [],
#     "var_dims": [],
#     "time_start": None,
#     "time_end": None,
#     "modified_date_time": None,
#     "collection_interval_time": [],
#     "collection_interval_time_final": [],
#     "item_bbox": [],
#     "item_footprint": None,
#     "collection_bbox": [],
#     "collection_footprint": None,
#     "collection_footprint_point": None,
#     "services": [],
#     "dataset": None,
#     "catalog_url": None,
#     "catalog_id": None,
# }


harvesting_vars = {  # type: ignore
    "item_id": None,
    "description": None,
    "horizontal_ids_lat": None,
    "horizontal_ids_lon": None,
    "horizontal_axis_x": None,
    "horizontal_axis_y": None,
    "horizontal_extent_lon_min": None,
    "horizontal_extent_lon_max": None,
    "horizontal_extent_lat_min": None,
    "horizontal_extent_lat_max": None,
    "horizontal_description_lon": None,
    "horizontal_description_lat": None,
    "horizontal_reference_system": None,
    "vertical_ids": None,
    "vertical_axis": None,
    "vertical_extent_upper": None,
    "vertical_extent_lower": None,
    "vertical_description": None,
    "temporal_id": None,
    "temporal_axis": None,
    "temporal_extent_start_datetime": None,
    "temporal_extent_end_datetime": None,
    "temporal_description": None,
    "variable_description": [],  # Description of each variable
    "variable_dimensions": [],  # dimention of each variable
    "variable_ids": [],  # Variable names
    "variable_unit": [],  # Variable units
    "variable_types": None,  # Variable types
    "services": [],
    "dataset": None,
    "catalog_url": None,
    "main_dataset_url": None,
    "catalog_id": None,
    "item_bbox": [],
    "item_footprint": None,
    "collection_bbox": [],
    "collection_footprint": None,
    "collection_footprint_point": None,
    "collection_interval_time": [],
    "modified_date_time": None,
    "collection_interval_time_final": [],
}

nested_scenarios = [
    "First Scenario",
    "Third Scenario",
    "Second Scenario",
    "Eighth Scenario",
    "Ninth Scenario",
]
none_nested_scenarios = [
    "Fourth Scenario",
    "Fifth Scenario",
    "Sixth Scenario",
    "Seventh Scenario",
]
supported_webservices = [
    "ncml",
    "wms",
    "iso",
    "dap4",
    "NCML",
    "WMS",
    "ISO",
    "DAP4",
    "catalog",
    "all",
]
webservice_properties_main_keys = ["webservice"]
empty_collection = pystac.Collection(
    id="an-empty-collection-id",
    description="An Empty collection description",
    extent=pystac.Extent(
        spatial=pystac.SpatialExtent(bboxes=[0.0, 0.0]),
        temporal=pystac.TemporalExtent(
            intervals=[[datetime.utcnow(), datetime.utcnow()]]
        ),
    ),
)



# # SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
# #
# # SPDX-License-Identifier: CC0-1.0

# import sys

# from tds2stac.main_app import Nested_Collections

# sys.path.append("../")
from tds2stac import NestedCollectionInspector


def test_first_case_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario1/test/catalog.html",
    )


def test_first_case_gt_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario1/test/catalog.html",
        nested_number=1,
    )


def test_second_case_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario2.html",
        nested_number=0,
    )


def test_second_case_gt_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario2.html",
        nested_number=1,
    )


def test_third_case_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario3/test/nested/first/catalog.html",
        nested_number=0,
    )


def test_third_case_gt_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario3/test/nested/first/catalog.html",
        nested_number=1,
    )


def test_eighth_case_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario8.html",
        nested_number=0,
    )


def test_eighth_case_gt_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario8.html",
        nested_number=1,
    )


def test_ninth_case_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario9.html",
        nested_number=0,
    )


def test_ninth_case_gt_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/scenario9.html",
        nested_number=1,
    )


def test_random_case_0():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/main.html",
        nested_number=0,
    )


def test_random_case_1():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/main.html",
        nested_number=1,
    )


def test_random_case_2():
    NestedCollectionInspector(
        "http://localhost:8080/thredds/catalog/catalogs/main.html",
        nested_number=2,
    )



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

import os
from urllib import parse as urlparse

import requests


# TODO: we need to think about data with different extesions than .nc
# like .tar in Musica. Also it would be greate to first split the
# name based on the extension and then starts changing the name.
def replacement_func(url):
    # TODO: fix the following issue when the user input wrong address like ncml address instead of data address:
    #   File "/Users/hadizadeh-m/dev/20231026/tds2stac/tds2stac/utils.py", line 18, in replacement_func
    #     temp = path_.split(k)
    # TypeError: a bytes-like object is required, not 'str'

    """A function for making a an id from catalog URL for collections"""
    splitted_arr = urlparse.urlsplit(url)
    trans_dict_path = {"%": "", "/catalog.xml": "", "/thredds/": "", "/": " "}
    path_ = splitted_arr.path
    query_ = splitted_arr.query
    for k, v in trans_dict_path.items():
        temp = path_.split(k)
        path_ = v.join(temp)
    if splitted_arr.query != "":
        trans_dict_query = {"=": " ", "?": " ", "/": " "}
        for k, v in trans_dict_query.items():
            temp = query_.split(k)
            query_ = v.join(temp)
        replaced_url = path_.replace(".xml", "") + " " + query_
        trans_dict_replaced = {".nc": "", "-": " ", "_": " "}
        for k, v in trans_dict_replaced.items():
            temp = replaced_url.split(k)
            replaced_url = v.join(temp)
    else:
        replaced_url = path_
        trans_dict_replaced = {".xml": "", ".nc": "", "-": " ", "_": " "}
        for k, v in trans_dict_replaced.items():
            temp = replaced_url.split(k)
            replaced_url = v.join(temp)
    replaced_url = replaced_url.replace(".", " ")
    return replaced_url.title()


def replacement_func_collection_item_id(url):
    # TODO: fix the following issue when the user input wrong address like ncml address instead of data address:
    #   File "/Users/hadizadeh-m/dev/20231026/tds2stac/tds2stac/utils.py", line 18, in replacement_func
    #     temp = path_.split(k)
    # TypeError: a bytes-like object is required, not 'str'

    """A function for making a an id from catalog URL for collections"""
    splitted_arr = urlparse.urlsplit(url)
    trans_dict_path = {"%": "", "/catalog.xml": "", "/thredds/": "", "/": " "}
    path_ = splitted_arr.path
    query_ = splitted_arr.query
    for k, v in trans_dict_path.items():
        temp = path_.split(k)
        path_ = v.join(temp)
    if splitted_arr.query != "":
        trans_dict_query = {"=": " ", "?": " ", "/": " "}
        for k, v in trans_dict_query.items():
            temp = query_.split(k)
            query_ = v.join(temp)
        replaced_url = path_.replace(".xml", "") + " " + query_
        trans_dict_replaced = {".nc": "", "-": " ", "_": " "}
        for k, v in trans_dict_replaced.items():
            temp = replaced_url.split(k)
            replaced_url = v.join(temp)
    else:
        replaced_url = path_
        trans_dict_replaced = {".xml": "", ".nc": "", "-": " ", "_": " "}
        for k, v in trans_dict_replaced.items():
            temp = replaced_url.split(k)
            replaced_url = v.join(temp)
    replaced_url = replaced_url.lower()
    replaced_url = replaced_url.replace(" ", "_")
    replaced_url = replaced_url.replace(".", "_")
    return replaced_url


def html2xml(url):
    """A function for making a an xml URL from html URL"""
    u = urlparse.urlsplit(url)
    path, extesion = os.path.splitext(u.path)
    if extesion == ".html":
        u = urlparse.urlsplit(url.replace(".html", ".xml"))
    return u.geturl()


def xml2html(url):
    """A function for making a an html URL from xml URL"""
    u = urlparse.urlsplit(url)
    path, extesion = os.path.splitext(u.path)
    if extesion == ".xml":
        u = urlparse.urlsplit(url.replace(".xml", ".html"))
    return u.geturl()


def get_xml(url, auth):
    """A function for getting XML content from url"""
    try:
        xml_url = requests.get(url, None, auth=auth, verify=False)
        xml = xml_url.text.encode("utf-8")
        return xml
    except BaseException:
        pass
        return None


def references_urls(url, additional):
    splitted_arr = urlparse.urlsplit(url)
    common_url = str(splitted_arr.scheme) + "://" + str(splitted_arr.netloc)
    wihtout_catalog_xml = urlparse.urljoin(
        common_url, os.path.split(splitted_arr.path)[0]
    )

    if not additional:
        final_url = url
    elif additional[:4] == "http":
        # finding http or https
        final_url = additional
    elif additional[0] == "/":
        # Absolute paths
        final_url = urlparse.urljoin(common_url, additional)
    else:
        # Relative paths.
        final_url = wihtout_catalog_xml + "/" + additional
    return final_url


def xml_processing(catalog, auth):
    """A function for getting out XML details of a catalog URL"""
    catalog_xml = html2xml(catalog)
    catalog_id = replacement_func(catalog_xml)
    xml_final = get_xml(catalog_xml, auth)
    return catalog_xml, catalog_id, xml_final


def xml_tag_name_ncml(input_xml, var_name):
    """A function for finding the tag names in NcML XML files"""

    # A list for recognizign the exceptions.
    # This list contains variable's name with same `input_xml.tag` and different `var_name`s
    exception_list = ["var_lists", "var_dims", "var_descs", "keyword"]

    if var_name in exception_list:
        return input_xml.tag + "_" + var_name
    else:
        return str(input_xml.get("name")) + "_" + var_name


def xml_tag_finder(input_xml, web_service, var_name):
    """A function for finding the tag names in all TDS webservices"""

    tag_finder_dict = {
        "iso": input_xml.tag + "_" + var_name,
        "ncml": xml_tag_name_ncml(input_xml, var_name),
        "wms": input_xml.tag + "_" + var_name,
    }

    return tag_finder_dict.get(
        web_service,
    )



# # SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
# #
# # SPDX-License-Identifier: CC0-1.0

# import sys

# from tds2stac.main_app import Catalog_summary

# sys.path.append("../")


# def test_catalog_summary_first_scenario_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/regional/seas5_bc/forecast_measures/catalog.html",
#         depth_number=0,
#     )


# def test_catalog_summary_first_scenario_gt_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/regional/seas5_bc/forecast_measures/catalog.html",
#         depth_number=1,
#     )


# def test_catalog_summary_second_scenario_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/sensor_catalog_ext.html",
#         depth_number=0,
#     )


# def test_catalog_summary_second_scenario_gt_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/sensor_catalog_ext.html",
#         depth_number=1,
#     )


# def test_catalog_summary_third_scenario_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/chirps/catalog.html",
#         depth_number=0,
#     )


# def test_catalog_summary_third_scenario_gt_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/chirps/catalog.html",
#         depth_number=1,
#     )


# def test_catalog_summary_fourth_scenario():
#     Catalog_summary(
#         "https://thredds.atmohub.kit.edu/thredds/catalog/snowfogs/catalog.html",
#         depth_number=0,
#     )


# def test_catalog_summary_fifth_scenario():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/climate/raster/global/chelsa/v1.2/catalog.html",
#         depth_number=1,
#     )


# def test_catalog_summary_sixth_scenario():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/daily/catalog.html?dataset=era5_sfc_0.25_single/daily/ERA5_daily_sp_1979.nc",
#         depth_number=0,
#     )


# def test_catalog_summary_seventh_scenario():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/climate_catalog_ext.html?dataset=gleam_v3.5a_daily_aggregated",
#         depth_number=1,
#     )


# def test_catalog_summary_eighth_scenario_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/transfer.html",
#         depth_number=0,
#     )


# def test_catalog_summary_eighth_scenario_gt_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/transfer.html",
#         depth_number=1,
#     )


# def test_catalog_summary_ninth_scenario_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/hydrogfd/v3.0/catalog.html",
#         depth_number=0,
#     )


# def test_catalog_summary_ninth_scenario_gt_zero():
#     Catalog_summary(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/hydrogfd/v3.0/catalog.html",
#         depth_number=1,
#     )



from datetime import datetime

# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
import pytz


class Temporal(object):
    def regulator(self, main_dict, temporal_format_by_dataname, data_name):
        errors = []
        if (
            main_dict["temporal_extent_start_datetime"] is not None
            and temporal_format_by_dataname is None
        ):
            main_dict["temporal_extent_start_datetime"] = main_dict[
                "temporal_extent_start_datetime"
            ]
        elif (
            main_dict["temporal_extent_start_datetime"] is None
            or main_dict["temporal_extent_start_datetime"] is not None
        ) and temporal_format_by_dataname is not None:
            print("data_name", data_name)
            print(
                "temporal_format_by_dataname",
                datetime.strptime(
                    str(data_name),
                    temporal_format_by_dataname,
                )
                .strftime("%Y-%m-%dT%H:%M:%SZ")
                .lstrip("0"),
            )
            main_dict["temporal_extent_start_datetime"] = (
                datetime.strptime(
                    str(data_name),
                    temporal_format_by_dataname,
                ).strftime("%Y-%m-%dT%H:%M:%SZ")
                # .lstrip("0")
                # .replace(" 0", " ")
            )
        elif (
            main_dict["temporal_extent_start_datetime"] is None
            and temporal_format_by_dataname is None
        ):
            errors.append(
                "Start DateTime is None in the given dataset. Please review your `tag_config.json` file or use `temporal_format_by_dataname` attr to the date of your data will be read from file names."
            )
        if (
            main_dict["temporal_extent_end_datetime"] is not None
            and temporal_format_by_dataname is None
        ):
            main_dict["temporal_extent_end_datetime"] = main_dict[
                "temporal_extent_end_datetime"
            ]
        elif (
            main_dict["temporal_extent_end_datetime"] is None
            or main_dict["temporal_extent_end_datetime"] is not None
        ) and temporal_format_by_dataname is not None:
            main_dict["temporal_extent_end_datetime"] = (
                datetime.strptime(
                    str(data_name),
                    temporal_format_by_dataname,
                ).strftime("%Y-%m-%dT%H:%M:%SZ")
                # .lstrip("0")
                # .replace(" 0", " ")
            )
        elif (
            main_dict["temporal_extent_end_datetime"] is None
            and temporal_format_by_dataname is None
        ):
            errors.append(
                "End DateTime is None in the given dataset. Please review your `tag_config.json` file or use `temporal_format_by_dataname` attr to the date of your data will be read from file names."
            )

        try:
            main_dict["collection_interval_time"].append(
                datetime.strptime(
                    main_dict["temporal_extent_start_datetime"],
                    "%Y-%m-%dT%H:%M:%S.%fZ",
                ).replace(tzinfo=pytz.utc)
            )
            main_dict["collection_interval_time"].append(
                datetime.strptime(
                    main_dict["temporal_extent_end_datetime"],
                    "%Y-%m-%dT%H:%M:%S.%fZ",
                ).replace(tzinfo=pytz.utc)
            )
        except ValueError:
            main_dict["collection_interval_time"].append(
                datetime.strptime(
                    main_dict["temporal_extent_start_datetime"],
                    "%Y-%m-%dT%H:%M:%SZ",
                ).replace(tzinfo=pytz.utc)
            )
            main_dict["collection_interval_time"].append(
                datetime.strptime(
                    main_dict["temporal_extent_end_datetime"],
                    "%Y-%m-%dT%H:%M:%SZ",
                ).replace(tzinfo=pytz.utc)
            )

        main_dict["collection_interval_time"] = sorted(
            main_dict["collection_interval_time"]
        )

        if errors != []:
            return errors



# # SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
# #
# # SPDX-License-Identifier: CC0-1.0

from tds2stac import Recognizer


def test_first_case():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario1/test/catalog.html",
        nested_check=True,
        logger_properties={"logger_handler": "FileHandler"},
    )


def test_second_case():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario2.html",
        nested_check=True,
        logger_properties={
            "logger_handler": "FileHandler",
            "logger_handler_filename": "test.log",
            "logger_handler_mode": "w",
        },
    )


def test_third_case():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario3/test/nested/first/catalog.html",
        nested_check=True,
        logger_properties={"logger_handler": "SMTPHandler"},
    )


def test_fourth_case1():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario4.html",
        logger_properties={"logger_handler": "HTTPHandler"},
    )


def test_fifth_case():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html"
    )


def test_sixth_case():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html?dataset=a5f5be12-e4de-4977-adb7-a06480f65f48/2020336.nc"
    )


def test_seventh_case():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario7.html?dataset=2786bf07-8e3b-44d3-82bd-ee27024751ec"
    )


def test_eighth_case():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario8/test/catalog.html",
        nested_check=True,
    )


def test_ninth_case():
    Recognizer(
        "http://localhost:8080/thredds/catalog/catalogs/scenario9.html",
        nested_check=True,
    )


# def test_first_case():
#     Recognizer(
#         "http://localhost:8080/thredds/catalog/catalogs/scenario1/test/catalog.html",
#         nested_check=True,
#         logger_properties={"logger_handler": "FileHandler"},
#     )


# def test_second_case():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/sensor_catalog_ext.html",
#         nested_check=True,
#         logger_properties={
#             "logger_handler": "FileHandler",
#             "logger_handler_filename": "test.log",
#             "logger_handler_mode": "w",
#         },
#     )


# def test_third_case():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/chirps/catalog.html",
#         nested_check=True,
#         logger_properties={"logger_handler": "SMTPHandler"},
#     )


# def test_fourth_case1():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/bio_geo_chem_catalog_ext.xml",
#         logger_properties={"logger_handler": "HTTPHandler"},
#     )


# def test_fifth_case():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/climate/raster/global/chelsa/v1.2/catalog.html"
#     )


# def test_sixth_case():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/daily/catalog.html?dataset=era5_sfc_0.25_single/daily/ERA5_daily_sp_1981.nc"
#     )


# def test_seventh_case():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/swabian_moses_2021.html?dataset=swabian_moses_aggregation"
#     )


# def test_eighth_case():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/transfer.html",
#         nested_check=True,
#     )


# def test_ninth_case():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/hydrogfd/v3.0/catalog.html",
#         nested_check=True,
#     )


# def test_case_finder():
#     Recognizer(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/reg_clim_sys_catalog_ext.html",
#         nested_check=True,
#     )



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

import json

# from ..statics import constants


class Verifier(object):  # type: ignore
    def webservice_properties(
        self,
        webservice_properties: dict,
    ):
        if webservice_properties is None:
            # webservice_properties["webservice"] = "all"
            webservice_properties["web_service_config_file"] = "default"
        # elif "webservice" not in webservice_properties.keys():
        #     # TODO: defining a warning, because webservice attribute is mandatory and you didnt define it there we will pu it as all
        #     webservice_properties["webservice"] = "all"
        # elif not any(webservice_properties["webservice"] in str(ws) for ws in constants.supported_webservices):
        #     # TODO: issue a warning that the webservice is not valid and we will put it as all
        #     webservice_properties["webservice"] = "all"
        elif "web_service_config_file" not in webservice_properties.keys():
            # TODO: adding another warning logger here
            webservice_properties["web_service_config_file"] = "default"
        elif "web_service_config_file" in webservice_properties.keys():
            try:
                json.load(
                    open(webservice_properties["web_service_config_file"])
                )
            except Exception:
                # TODO: adding another warning logger here
                print(
                    "The json file is not valid. Check the json file address and try again. For this case we set the `web_service_config_file` to `default`"
                )
                webservice_properties["web_service_config_file"] = "default"

    def extension_properties(
        self,
        extension_properties: dict,
    ):
        List_of_extension_keys = [
            "item_datacube",
            "item_common_metadata",
            "item_scientific",
            "collection_scientific",
            "custom_extension_name",
            "custom_extension_script",
        ]
        custom_extension_keys = [
            "custom_extension_name",
            "custom_extension_script",
        ]
        if extension_properties is None or extension_properties == {}:
            extension_properties["item_datacube"] = False
            extension_properties["item_common_metadata"] = False
            extension_properties["item_scientific"] = False
            extension_properties["collection_scientific"] = False
            extension_properties["custom_extension_name"] = None
            extension_properties["custom_extension_script"] = None

        elif extension_properties is not None or extension_properties != {}:
            for i in List_of_extension_keys:
                if i not in extension_properties.keys():
                    if i not in custom_extension_keys:
                        extension_properties[i] = False
                    else:
                        extension_properties[i] = None

    def asset_properties(
        self,
        asset_properties: dict,
    ):
        asset_properties_list = [
            "item_thumbnail",
            "item_overview",
            "item_getminmax_thumbnail",
            "collection_thumbnail",
            "collection_overview",
            "collection_thumbnail_link",
            "item_assets_list_allowed",
            "item_assets_list_avoided",
            "collection_assets_list_allowed",
            "collection_assets_list_avoided",
            "item_scientific",
            "explore_data",
            "check_get_metadata",
            "jupyter_notebook",
        ]
        none_false_asset_properties = [
            "collection_thumbnail_link",
            "collection_overview",
            "collection_thumbnail",
            "item_assets_list_allowed",
            "item_assets_list_avoided",
            "collection_assets_list_allowed",
            "collection_assets_list_avoided",
        ]
        if asset_properties is None:
            asset_properties["item_thumbnail"] = False
            asset_properties["item_overview"] = False
            asset_properties["item_getminmax_thumbnail"] = False
            asset_properties["collection_thumbnail"] = None
            asset_properties["collection_overview"] = None
            asset_properties["collection_thumbnail_link"] = None
            asset_properties["item_scientific"] = False
            asset_properties["item_assets_list_allowed"] = None
            asset_properties["item_assets_list_avoided"] = None
            asset_properties["collection_assets_list_allowed"] = None
            asset_properties["collection_assets_list_avoided"] = None
            asset_properties["explore_data"] = False
            asset_properties["check_get_metadata"] = False
            asset_properties["jupyter_notebook"] = False

        elif asset_properties is not None:
            for i in asset_properties_list:
                if i not in asset_properties.keys():
                    if i not in none_false_asset_properties:
                        asset_properties[i] = False
                    else:
                        asset_properties[i] = None

    def logger_properties(
        self,
        logger_properties: dict,
    ) -> dict:
        if logger_properties == {}:
            logger_properties["logger_handler"] = "NullHandler"
        return logger_properties



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: EUPL-1.2

"""Test file for imports."""


def test_package_import():
    """Test the import of the main package."""
    import tds2stac  # noqa: F401


# # nested depth = 0
# def test_cc():
#     TDS2STACIntegrator(
#         "http://localhost:8080/thredds/catalog/testAll/catalog.html?dataset=testDatasetScan/2004050412_eta_211.nc",
#         stac_dir="stac/",
#     )



# # SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
# #
# # SPDX-License-Identifier: CC0-1.0

# from tds2stac.main_app import Web_Services

# url = [
#     "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/climatology/catalog.html",
#     "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/climatology/catalog.html",
#     "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/catalog.html",
# ]


# def test_web_service():
#     for i in url:
#         Web_Services(i)



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
from tds2stac import TDS2STACIntegrator


# nested depth = 0
def test_tds2stacintegrator_nested_depth_zero_all_possible_extensions():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=0,
        extension_properties={
            "item_extensions": [
                "common_metadata",
                "item_datacube_extension",
                (
                    "scientific_extension",
                    "item",
                    "./custom_based_on_defined_extension_pystac.py",
                ),
                (
                    "contact_extension",
                    "item",
                    "./custom_based_on_none_defined_extension_pystac.py",
                ),
            ],
        },
    )


def test_tds2stacintegrator_nested_depth_zero_with_webservice():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        datetime_filter=["2010-02-18T00:00:00Z", "2010-02-18T00:00:00Z"],
        depth_number=0,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
    )


def test_tds2stacintegrator_nested_depth_zero_without_webservice():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=0,
    )


def test_tds2stacintegrator_nested_depth_zero_with_asset():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html?dataset=a5f5be12-e4de-4977-adb7-a06480f65f48/2020336.nc",
        stac_dir="stac/",
        limited_number=2,
        depth_number=0,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
        extension_properties={
            "item_extensions": [
                "common_metadata",
                "item_datacube_extension",
                (
                    "scientific_extension",
                    "item",
                    "./custom_based_on_defined_extension_pystac.py",
                ),
                (
                    "contact_extension",
                    "item",
                    "./nn/custom_based_on_none_defined_extension_pystac.py",
                ),
            ],
        },
        asset_properties={
            "item_thumbnail": True,
            "item_getminmax_thumbnail": True,
            "explore_data": True,
            "verify_explore_data": True,
            "jupyter_notebook": True,
            "collection_thumbnail": "link",
            "collection_overview": "wms",
            "collection_link": "https://images.fineartamerica.com/images-medium-large-5/global-precipitation-william-putmannasa-goddard-space-flight-center.jpg",
            "assets_list_allowed": ["wms", "wcs", "wfs"],
            "assets_list_avoided": ["http"],
        },
    )


# nested depth gt zero
def test_tds2stacintegrator_nested_depth_gt_zero_all_possible_extensions():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=1,
        extension_properties={
            "item_extensions": [
                "common_metadata",
                "item_datacube_extension",
                (
                    "scientific_extension",
                    "item",
                    "./custom_based_on_defined_extension_pystac.py",
                ),
                (
                    "contact_extension",
                    "item",
                    "./custom_based_on_none_defined_extension_pystac.py",
                ),
            ],
        },
    )


def test_tds2stacintegrator_nested_depth_gt_zero_with_webservice():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=1,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
    )


def test_tds2stacintegrator_nested_depth_gt_zero_without_webservice():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=1,
    )


def test_tds2stacintegrator_nested_depth_gt_zero_with_asset():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=1,
        depth_number=1,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
        extension_properties={
            "item_extensions": [
                "common_metadata",
                "item_datacube_extension",
                (
                    "scientific_extension",
                    "item",
                    "./custom_based_on_defined_extension_pystac.py",
                ),
                (
                    "contact_extension",
                    "item",
                    "./nn/custom_based_on_none_defined_extension_pystac.py",
                ),
            ],
        },
        asset_properties={
            "item_thumbnail": True,
            "item_overview": True,
            "item_getminmax_thumbnail": True,
            "explore_data": True,
            "verify_explore_data": True,
            "jupyter_notebook": True,
            "collection_thumbnail": "link",
            "collection_overview": "wms",
            "collection_link": "https://images.fineartamerica.com/images-medium-large-5/global-precipitation-william-putmannasa-goddard-space-flight-center.jpg",
            "assets_list_allowed": ["wms", "wcs", "wfs"],
            "assets_list_avoided": ["http"],
        },
    )


# Another case
def test_tds2stacintegrator_nested_depth_gt_zero_all_possible_extensions1():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario3/test/nested/first/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=1,
        extension_properties={
            "item_extensions": [
                "common_metadata",
                "item_datacube_extension",
                (
                    "scientific_extension",
                    "item",
                    "./custom_based_on_defined_extension_pystac.py",
                ),
                (
                    "contact_extension",
                    "item",
                    "./custom_based_on_none_defined_extension_pystac.py",
                ),
            ],
        },
    )


def test_tds2stacintegrator_nested_depth_gt_zero_with_webservice1():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario3/test/nested/first/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=1,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
        logger_properties={"logger_handler": "StreamHandler"},
    )


def test_tds2stacintegrator_nested_depth_gt_zero_without_webservice1():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario3/test/nested/first/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=1,
        logger_properties={"logger_handler": "StreamHandler"},
    )


def test_tds2stacintegrator_nested_depth_gt_zero_with_asset1():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario3/test/catalog.html",
        stac_dir="stac/",
        limited_number=1,
        depth_number=1,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
        extension_properties={
            "item_extensions": [
                "common_metadata",
                "item_datacube_extension",
                (
                    "scientific_extension",
                    "item",
                    "./custom_based_on_defined_extension_pystac.py",
                ),
                (
                    "contact_extension",
                    "item",
                    "./custom_based_on_none_defined_extension_pystac.py",
                ),
            ],
        },
        asset_properties={
            # "item_thumbnail": True,
            # "item_getminmax_thumbnail": True,
            "explore_data": True,
            "verify_explore_data": True,
            "jupyter_notebook": True,
            "collection_thumbnail": "link",
            "collection_link": "https://images.fineartamerica.com/images-medium-large-5/global-precipitation-william-putmannasa-goddard-space-flight-center.jpg",
            "assets_list_allowed": ["wms", "wcs", "wfs"],
            "assets_list_avoided": ["http"],
            # "item_extra_custom_assets": [("extra_item_asset", "item_asset"),],
            # "extra_custom_assets":[{"href": "", "title": "", "description": "", "media_type": "", "roles": "", "extra_fields": ""},],
        },
        logger_properties={"logger_handler": "StreamHandler"},
    )


# none nested depth = 0
def test_tds2stacintegrator_none_nested_depth_zero_all_possible_extensions():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=1,
        depth_number=0,
        extension_properties={
            "item_extensions": [
                "common_metadata",
                "item_datacube_extension",
                (
                    "scientific_extension",
                    "item",
                    "./custom_based_on_defined_extension_pystac.py",
                ),
                (
                    "contact_extension",
                    "item",
                    "./custom_based_on_none_defined_extension_pystac.py",
                ),
            ],
        },
        logger_properties={"logger_handler": "StreamHandler"},
    )


def test_tds2stacintegrator_none_nested_depth_zero_with_webservice():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=0,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
        logger_properties={"logger_handler": "StreamHandler"},
    )


def test_tds2stacintegrator_none_nested_depth_zero_without_webservice():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        stac_dir="stac/",
        limited_number=2,
        depth_number=0,
        logger_properties={"logger_handler": "StreamHandler"},
    )


# def test_tds2stacintegrator_none_nested_depth_zero_with_asset():
#     TDS2STACIntegrator(
#         "http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html?dataset=a5f5be12-e4de-4977-adb7-a06480f65f48/2020336.nc",
#         stac_dir="stac/",
#         limited_number=1,
#         depth_number=0,
#         webservice_properties={
#             "web_service_config_file": "./tag_example.json",
#         },
#         extension_properties={
#             "item_extensions": [
#                 "common_metadata",
#                 "item_datacube_extension",
#                 (
#                     "scientific_extension",
#                     "item",
#                     "./custom_based_on_defined_extension_pystac.py",
#                 ),
#                 (
#                     "contact_extension",
#                     "item",
#                     "./custom_based_on_none_defined_extension_pystac.py",
#                 ),
#             ],
#         },
#         asset_properties={
#             "item_thumbnail": True,
#             "item_getminmax_thumbnail": True,
#             "explore_data": True,
#             "verify_explore_data": True,
#             "jupyter_notebook": True,
#             "collection_thumbnail": "wms",
#             "assets_list_allowed": ["wms", "wcs", "wfs"],
#             "assets_list_avoided": ["http"],
#             # "item_extra_custom_assets": [("extra_item_asset", "item_asset"),],
#             # "extra_custom_assets":[{"href": "", "title": "", "description": "", "media_type": "", "roles": "", "extra_fields": ""},],
#         },
#         logger_properties={"logger_handler": "NullHandler"},
#     )


def test_tds2stacintegrator_atmohub_linestring():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/main.html?dataset=f3627295-28e0-48a2-b9b5-c2bdac5fb2cf",
        stac_dir="stac/",
        stac_existance=True,
        stac_existance_collection=True,
        item_geometry_linestring=True,
        limited_number=2,
        depth_number=0,
        webservice_properties={
            "webservice": "all",
            "web_service_config_file": "./tag_example.json",
        },
        logger_properties={"logger_handler": "StreamHandler"},
    )


def test_tds2stacintegrator_atmohub_point():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/main.html?dataset=a964b0f6-321e-481e-a43b-f036dd333ae2",
        stac_dir="stac/",
        stac_existance=True,
        stac_existance_collection=True,
        limited_number=2,
        depth_number=0,
        webservice_properties={
            "webservice": "all",
            "web_service_config_file": "./tag_example.json",
        },
        logger_properties={"logger_handler": "StreamHandler"},
    )


# TODO: fix this test
def test_tds2stacintegrator_atmohub_spatial_information():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/main.html?dataset=a964b0f6-321e-481e-a43b-f036dd333ae2",
        stac_dir="stac/",
        stac_existance=True,
        stac_existance_collection=True,
        item_geometry_linestring=True,
        spatial_information=[0, 0, 0, 0],
        limited_number=2,
        depth_number=0,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
        logger_properties={"logger_handler": "StreamHandler"},
    )


def test_tds2stacintegrator_agg():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario7.html?dataset=2786bf07-8e3b-44d3-82bd-ee27024751ec",
        stac_dir="stac/",
        stac_existance=True,
        stac_existance_collection=True,
        aggregated_dataset_url="https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/regional/eobs/v27.0e/catalog.html",
        limited_number=2,
        collection_tuples=[
            (
                "catalog_catalogues_climate_catalog_ext_html_dataset_eobs_daily_0_1_aggregated_v27_aggregated",
                "sfsvs",
                "sadsdsdf",
                "dassdsdsfd",
            )
        ],
        depth_number=0,
        webservice_properties={
            "web_service_config_file": "./tag_example.json",
        },
        logger_properties={"logger_handler": "StreamHandler"},
    )


def test_tds2stacintegrator_empty():
    TDS2STACIntegrator(
        "http://localhost:8080/thredds/catalog/catalogs/scenario4.html",
        stac_dir="stac/",
        collection_tuples=[
            (
                "catalog_catalogues_bio_geo_chem_catalog_ext_html_empty",
                "sfsvs",
                "sadsdsdf",
                "dassdsdsfd",
            )
        ],
        logger_properties={"logger_handler": "StreamHandler"},
    )


# TODO: fix this test
def test_tds2stacintegrator_error():
    TDS2STACIntegrator(
        "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/atmos_obs/lidar/Data/20220714/catalog.html?dataset=lidar/Data/20220714/e2271407.590015",
        stac_dir="stac/",
        logger_properties={"logger_handler": "StreamHandler"},
    )


# TODO: fix this test
def test_tds2stacintegrator_no_spatial_no_temporal():
    TDS2STACIntegrator(
        "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/regional/seas5_bc/forecast_measures/raster/catalog.html",
        stac_dir="stac/",
        limited_number=1,
        logger_properties={"logger_handler": "StreamHandler"},
    )



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
from datetime import datetime

import pytz


class CommonMetadata:
    def item(self, item, harvesting_vars):
        # Add auxiliary information to items
        try:
            item.common_metadata.start_datetime = datetime.strptime(
                harvesting_vars["temporal_extent_start_datetime"],
                "%Y-%m-%dT%H:%M:%S.%fZ",
            ).replace(tzinfo=pytz.utc)
        except ValueError:
            item.common_metadata.start_datetime = datetime.strptime(
                harvesting_vars["temporal_extent_start_datetime"],
                "%Y-%m-%dT%H:%M:%SZ",
            ).replace(tzinfo=pytz.utc)
        try:
            item.common_metadata.end_datetime = datetime.strptime(
                harvesting_vars["temporal_extent_end_datetime"],
                "%Y-%m-%dT%H:%M:%S.%fZ",
            ).replace(tzinfo=pytz.utc)
        except ValueError:
            item.common_metadata.end_datetime = datetime.strptime(
                harvesting_vars["temporal_extent_end_datetime"],
                "%Y-%m-%dT%H:%M:%SZ",
            ).replace(tzinfo=pytz.utc)
        if (
            harvesting_vars["description"] is not None
            and harvesting_vars["service_url_html"] is not None
        ):
            item.common_metadata.description = (
                harvesting_vars["description"]
                + "\n\n [Link to the data catalog in TDS]("
                + harvesting_vars["service_url_html"]
                + ")"
            )
        if harvesting_vars.get("constellation") is not None:
            item.common_metadata.constellation = harvesting_vars[
                "constellation"
            ]
        if harvesting_vars.get("gsd") is not None:
            item.common_metadata.gsd = harvesting_vars["gsd"]
        if harvesting_vars.get("instruments") is not None:
            item.common_metadata.instruments = harvesting_vars["instruments"]
        if harvesting_vars.get("license") is not None:
            item.common_metadata.license = harvesting_vars["license"]
        if harvesting_vars.get("mission") is not None:
            item.common_metadata.mission = harvesting_vars["mission"]
        if harvesting_vars.get("providers") is not None:
            item.common_metadata.providers = harvesting_vars["providers"]
        if harvesting_vars.get("title") is not None:
            item.common_metadata.title = harvesting_vars["title"]
        if harvesting_vars.get("updated") is not None:
            item.common_metadata.updated = harvesting_vars["updated"]
        if harvesting_vars.get("created") is not None:
            item.common_metadata.created = harvesting_vars["created"]



# # SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
# #
# # SPDX-License-Identifier: CC0-1.0

# import os
# import shutil

# from tds2stac.main_app import Existance_checker


# def test_existance_checker_true():
#     current_dir = os.getcwd()
#     if not os.path.exists(current_dir + "/stac/"):
#         os.mkdir(current_dir + "/stac/")
#     os.chdir(current_dir + "/stac/")

#     f = open("catalog.json", "w")
#     f.write("Now the file has more content!")
#     f.close()
#     Existance_checker(current_dir)
#     shutil.rmtree(current_dir + "/stac/")


# def test_existance_checker_false():
#     Existance_checker("/Users/PycharmProjects/tds2stac/")



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0


from typing import Union
from urllib import parse as urlparse
from urllib.parse import quote_plus

import pystac
import requests

from . import utils
from .statics import constants
from .thumbnails import Thumbnails


class Assets:
    """
    This class is tasked with the responsibility of incorporating
    assets into STAC-Collections and STAC-Items.
    """

    def item(
        self,
        harvesting_vars: dict,
        item: pystac.Item,
        url: str,
        Recognizer_output: Union[str, None] = None,
        aggregated_dataset_url: Union[str, None] = None,
        asset_properties: Union[dict, None] = dict(),
        logger_properties: dict = dict(),
    ):
        """
        This is a function for adding assets to STAC-Items.

        Args:
            harvesting_vars (dict): A dictionary containing the variables required for harvesting.
            item (pystac.Item): A STAC-Item.
            Recognizer_output (dict): A dictionary containing the senario output of the :class:`~tds2stac.Recognizer`.
            aggregated_dataset_url (str): The URL of the aggregated dataset.
            url (str): The URL of the catalog. It will be used for thumbnails
            asset_properties (dict): A dictionary containing the properties of the assets for more information refere to :class:`~tds2stac.TDS2STACIntegrator.asset_properties`.
            logger_properties (dict): A dictionary containing the properties of the logger for more information refere to :class:`~tds2stac.TDS2STACIntegrator.logger_properties`.
        """
        # Adding web services as assets into items
        if logger_properties is not None:
            self.logger_properties = logger_properties
        media_type_: Union[str, pystac.MediaType] = ""
        for service in harvesting_vars["services"]:
            if (
                service.get("serviceType") == "Compound"
                or service.get("serviceType") == "compound"
            ):
                if (
                    asset_properties is not None
                    and asset_properties.get("jupyter_notebook") is True
                ):
                    catalog_path = urlparse.urlsplit(url).path.replace(
                        "/thredds/catalog/", ""
                    )
                    additional = "/thredds/notebook/"
                    service_url_notebook = utils.references_urls(
                        url, additional
                    ) + harvesting_vars["dataset"].get("ID")
                    item.add_asset(
                        key="jupyter_notebook",
                        asset=pystac.Asset(
                            href=service_url_notebook
                            + "?catalog=%s&filename=%s"
                            % (catalog_path, "default_viewer.ipynb"),
                            title="Jupyter Notebook",
                            media_type=pystac.MediaType.HTML,
                        ),
                    )
                for s in service.findall("{%s}service" % constants.unidata):
                    service_url = utils.references_urls(
                        url, s.get("base")
                    ) + harvesting_vars["dataset"].get("urlPath")

                    if s.get("suffix") is not None:
                        service_url += s.get("suffix")
                    if s.get("name") in [
                        "iso",
                        "ncml",
                        "uddc",
                    ]:
                        service_url += "?dataset=%s&&catalog=%s" % (
                            harvesting_vars["catalog_id"],
                            quote_plus(harvesting_vars["catalog_url"]),
                        )
                    elif s.get("name") in [
                        "wms",
                    ]:
                        if (
                            asset_properties is not None
                            and asset_properties.get("explore_data") is True
                            and asset_properties.get("verify_explore_data")
                            is not True
                        ):
                            item.add_asset(
                                key="explore_data",
                                asset=pystac.Asset(
                                    href=utils.references_urls(
                                        url, "/thredds/Godiva.html"
                                    )
                                    + "?server="
                                    + service_url,
                                    title="Explore Data",
                                    media_type=pystac.MediaType.HTML,
                                ),
                            )
                        elif (
                            asset_properties is not None
                            and asset_properties.get("explore_data") is True
                            and asset_properties.get("verify_explore_data")
                            is True
                        ):
                            head = requests.head(
                                service_url + "?request=GetMetadata&item=menu"
                            )
                            if head.status_code == 200:
                                item.add_asset(
                                    key="explore_data",
                                    asset=pystac.Asset(
                                        href=utils.references_urls(
                                            url, "/thredds/Godiva.html"
                                        )
                                        + "?server="
                                        + service_url,
                                        title="Explore Data",
                                        media_type=pystac.MediaType.HTML,
                                    ),
                                )
                        service_url += "?service=WMS&version=1.3.0&request=GetCapabilities"
                    elif (
                        s.get("name")
                        in [
                            "http",
                        ]
                        and "Seventh Scenario" in str(Recognizer_output)
                        and aggregated_dataset_url is not None
                    ):
                        service_url = aggregated_dataset_url
                        media_type_ = pystac.MediaType.HTML
                    elif (
                        s.get("name")
                        in [
                            "http",
                        ]
                        and "Seventh Scenario" in str(Recognizer_output)
                        and aggregated_dataset_url is None
                    ):
                        # service_url += "?service=WMS&version=1.3.0&request=GetCapabilities"
                        media_type_ = pystac.MediaType.HTML

                    if s.get("name") in ["odap"]:
                        service_url += ".html"
                    # Determinatio of Media Type
                    if s.get("name") in [
                        "iso",
                        "ncml",
                        "wms",
                        "wcs",
                        "wfs",
                        "sos",
                    ]:
                        media_type_ = pystac.MediaType.XML
                    elif s.get("name") in [
                        "http"
                    ] and "Seventh Scenario" not in str(Recognizer_output):
                        media_type_ = "application/netcdf"
                    elif s.get("name") in [
                        "dap4",
                        "odap",
                        "uddc",
                    ]:
                        media_type_ = pystac.MediaType.HTML
                    else:
                        media_type_ = pystac.MediaType.TEXT

                    if asset_properties is None or (
                        asset_properties.get("assets_list_allowed") is None
                        and asset_properties.get("assets_list_avoided") is None
                    ):
                        item.add_asset(
                            key=s.get("name"),
                            asset=pystac.Asset(
                                href=service_url,
                                # title=without_slash,
                                media_type=media_type_,
                            ),
                        )

                    elif (
                        asset_properties is not None
                        and asset_properties["assets_list_allowed"] is not None
                        and asset_properties["assets_list_avoided"] is not None
                        and s.get("name")
                        in asset_properties["assets_list_allowed"]
                        and s.get("name")
                        not in asset_properties["assets_list_avoided"]
                    ):
                        item.add_asset(
                            key=s.get("name"),
                            asset=pystac.Asset(
                                href=service_url,
                                # title=without_slash,
                                media_type=media_type_,
                            ),
                        )
                    if asset_properties is not None:
                        if asset_properties.get("item_thumbnail") is not None:
                            thumbnail = Thumbnails()
                            thumbnail.item(
                                service,
                                harvesting_vars["dataset"],
                                harvesting_vars,
                                url,
                                item,
                                asset_properties["item_thumbnail"],
                                asset_properties["item_overview"],
                                asset_properties["item_getminmax_thumbnail"],
                                logger_properties=self.logger_properties,
                            )

    def collection(
        self,
        harvesting_vars: dict,
        collection_dict: dict,
        stac_catalog: pystac.Catalog,
        asset_properties: dict = dict(),
        logger_properties: dict = dict(),
    ):
        """
        This is a function for adding assets to STAC-Collections.

        Args:
            asset_properties (dict): A dictionary containing the properties of the assets for more information refere to :class:`~tds2stac.TDS2STACIntegrator.asset_properties`.
            harvesting_vars (dict): A dictionary containing the variables required for harvesting.
            collection_dict (dict): A dictionary containing the properties of the collection.
            stac_catalog (pystac.Catalog): A STAC-Catalog.
            logger_properties (dict): A dictionary containing the properties of the logger for more information refere to :class:`~tds2stac.TDS2STACIntegrator.logger_properties`.


        """
        if logger_properties is not None:
            self.logger_properties = logger_properties
        thumbnail = Thumbnails()
        thumbnail.collection(
            asset_properties["collection_thumbnail"],
            asset_properties["collection_overview"],
            dict(harvesting_vars)["services"],
            dict(harvesting_vars)["dataset"],
            dict(harvesting_vars),
            dict(collection_dict)["collection_id"],
            dict(harvesting_vars)["main_dataset_url"],
            stac_catalog,
            asset_properties["collection_link"],
            logger_properties=self.logger_properties,
        )



# # SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
# #
# # SPDX-License-Identifier: CC0-1.0

# import sys

# from tds2stac.main_app import Harvester

# sys.path.append("../")


# def test_harvester_first_case_iso_without_lat_long_time():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/regional/seas5_bc/forecast_measures/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         web_service="iso",
#         limited_number=10,
#     )


# def test_harvester_first_case_iso_wit_manually_lat_long_time():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/regional/seas5_bc/forecast_measures/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         web_service="iso",
#         limited_number=10,
#         spatial_information=[0, 0, 0, 0],
#     )


# def test_harvester_first_case_iso():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         web_service="iso",
#         limited_number=10,
#     )


# def test_harvester_first_case_ncml():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         web_service="ncml",
#         limited_number=10,
#     )


# def test_harvester_first_case_wms():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         web_service="wms",
#         limited_number=10,
#     )


# def test_harvester_second_case():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/sensor_catalog_ext.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#     )


# def test_harvester_third_case():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/chirps/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#     )


# def test_harvester_fourth_case():
#     Harvester(
#         "https://thredds.atmohub.kit.edu/thredds/catalog/snowfogs/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#     )


# def test_harvester_fifth_case():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/climate/raster/global/chelsa/v1.2/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#     )


# def test_harvester_fifth_case_collection_thubmnail():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/climate/raster/global/chelsa/v1.2/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#         collection_thumbnail=True,
#     )


# def test_harvester_fifth_case_item_thubmnail():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/climate/raster/global/chelsa/v1.2/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#         item_thumbnail=True,
#     )


# def test_harvester_fifth_case_item_getminmax_thumbnail():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/climate/raster/global/chelsa/v1.2/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#         item_thumbnail=True,
#         item_getminmax_thumbnail=True,
#     )


# def test_harvester_sixth_case():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/era5/sfc/single/daily/catalog.html?dataset=era5_sfc_0.25_single/daily/ERA5_daily_sp_1979.nc",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#     )


# def test_harvester_seventh_case():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/climate_catalog_ext.html?dataset=gleam_v3.5a_daily_aggregated",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#     )


# def test_harvester_seventh_case_with_url():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/climate_catalog_ext.html?dataset=gleam_v3.5a_daily_aggregated",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         aggregated_dataset_url="https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/gleam/v3.5a/catalog.html",
#     )


# def test_harvester_eighth_case():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/catalogues/transfer.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#     )


# def test_harvester_ninth_case():
#     Harvester(
#         "https://thredds.imk-ifu.kit.edu/thredds/catalog/regclim/raster/global/hydrogfd/v3.0/catalog.html",
#         stac=True,
#         stac_id="id",
#         stac_description="description",
#         stac_dir="/Users/hadizadeh-m/stac/",
#         limited_number=10,
#     )



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0


from shapely import geometry


class Spatial(object):
    def regulator(self, main_dict, spatial_information):
        """A function for regulating the spatial information of a catalog"""
        # TODO: make a functio to to the same procedure once for all

        errors = []
        if (
            main_dict["horizontal_extent_lon_min"] is not None
            and spatial_information is None
        ):
            main_dict["horizontal_extent_lon_min"] = main_dict[
                "horizontal_extent_lon_min"
            ].replace(",", ".")
        elif (
            main_dict["horizontal_extent_lon_min"] is None
            or main_dict["horizontal_extent_lon_min"] is not None
        ) and spatial_information is not None:
            if len(spatial_information) == 4:
                main_dict["horizontal_extent_lon_min"] = str(
                    spatial_information[0]
                )
            elif len(spatial_information) == 2:
                main_dict["horizontal_extent_lon_min"] = str(
                    spatial_information[0]
                )
            else:
                errors.append(
                    "There are inaccuracies in the `spatial_information` input values. Please include [horizontal_extent_lon_min, horizontal_extent_lon_max, horizontal_extent_lat_min, horizontal_extent_lat_max] or [long, lat] in the `spatial_information` attr, depending on the data's dimensions"
                )
        elif (
            main_dict["horizontal_extent_lon_min"] is None
            and spatial_information is None
        ):
            errors.append(
                "Minumum Longitude is None in the given dataset. Please review your `tag_config.json` file or use `spatial_information` attr to add the spatial coordinte"
            )

        if (
            main_dict["horizontal_extent_lon_max"] is not None
            and spatial_information is None
        ):
            main_dict["horizontal_extent_lon_max"] = main_dict[
                "horizontal_extent_lon_max"
            ].replace(",", ".")
        elif (
            main_dict["horizontal_extent_lon_max"] is None
            or main_dict["horizontal_extent_lon_max"] is not None
        ) and spatial_information is not None:
            if len(spatial_information) == 4:
                main_dict["horizontal_extent_lon_max"] = str(
                    spatial_information[1]
                )
            elif len(spatial_information) == 2:
                main_dict["horizontal_extent_lon_max"] = str(
                    spatial_information[0]
                )
            else:
                errors.append(
                    "There are inaccuracies in the `spatial_information` input values. Please include [horizontal_extent_lon_min, horizontal_extent_lon_max, horizontal_extent_lat_min, horizontal_extent_lat_max] or [long, lat] in the `spatial_information` attr, depending on the data's dimensions"
                )
        elif (
            main_dict["horizontal_extent_lon_max"] is None
            and spatial_information is None
        ):
            errors.append(
                "Maximum Longitude is None in the given dataset. Please review your `tag_config.json` file or use `spatial_information` attr to add the spatial coordinte"
            )

        if (
            main_dict["horizontal_extent_lat_min"] is not None
            and spatial_information is None
        ):
            main_dict["horizontal_extent_lat_min"] = main_dict[
                "horizontal_extent_lat_min"
            ].replace(",", ".")
        elif (
            main_dict["horizontal_extent_lat_min"] is None
            or main_dict["horizontal_extent_lat_min"] is not None
        ) and spatial_information is not None:
            if len(spatial_information) == 4:
                main_dict["horizontal_extent_lat_min"] = str(
                    spatial_information[2]
                )
            elif len(spatial_information) == 2:
                main_dict["horizontal_extent_lat_min"] = str(
                    spatial_information[1]
                )
            else:
                errors.append(
                    "There are inaccuracies in the `spatial_information` input values. Please include [horizontal_extent_lon_min, horizontal_extent_lon_max, horizontal_extent_lat_min, horizontal_extent_lat_max] or [long, lat] in the `spatial_information` attr, depending on the data's dimensions"
                )
        elif (
            main_dict["horizontal_extent_lat_min"] is None
            and spatial_information is None
        ):
            errors.append(
                "Minimum Latitude is None in the given dataset. Please review your `tag_config.json` file or use `spatial_information` attr to add the spatial coordinte"
            )

        if (
            main_dict["horizontal_extent_lat_max"] is not None
            and spatial_information is None
        ):
            main_dict["horizontal_extent_lat_max"] = main_dict[
                "horizontal_extent_lat_max"
            ].replace(",", ".")
        elif (
            main_dict["horizontal_extent_lat_max"] is None
            or main_dict["horizontal_extent_lat_max"] is not None
        ) and spatial_information is not None:
            if len(spatial_information) == 4:
                main_dict["horizontal_extent_lat_max"] = str(
                    spatial_information[3]
                )
            elif len(spatial_information) == 2:
                main_dict["horizontal_extent_lat_max"] = str(
                    spatial_information[1]
                )
            else:
                errors.append(
                    "There are inaccuracies in the `spatial_information` input values. Please include [horizontal_extent_lon_min, horizontal_extent_lon_max, horizontal_extent_lat_min, horizontal_extent_lat_max] or [long, lat] in the `spatial_information` attr, depending on the data's dimensions"
                )
        elif (
            main_dict["horizontal_extent_lat_max"] is None
            and spatial_information is None
        ):
            errors.append(
                "Maximum Latitude is None in the given dataset. Please review your `tag_config.json` file or use `spatial_information` attr to add the spatial coordinte"
            )
        if errors == []:
            # A condition for longitudes more than 180 e.g. 360 degree. Cause STAC doesn't support longs
            # more than 180
            if (
                float(main_dict["horizontal_extent_lon_min"]) > 180
                or float(main_dict["horizontal_extent_lon_max"]) > 180
            ):
                main_dict["horizontal_extent_lon_min"] = str(
                    float(main_dict["horizontal_extent_lon_min"]) - 180
                )
                main_dict["horizontal_extent_lon_max"] = str(
                    float(main_dict["horizontal_extent_lon_max"]) - 180
                )
            return (
                main_dict["horizontal_extent_lon_min"],
                main_dict["horizontal_extent_lon_max"],
                main_dict["horizontal_extent_lat_min"],
                main_dict["horizontal_extent_lat_max"],
            )
        else:
            return errors

    def harvester(self, main_dict, linestring=None):
        # TODO: for making the foorprint we have to check another attribute for making PolyLINE adn PolyGON with more than boxes points
        if (
            main_dict["horizontal_extent_lon_min"]
            == main_dict["horizontal_extent_lon_max"]
            or main_dict["horizontal_extent_lat_max"]
            == main_dict["horizontal_extent_lat_min"]
        ):
            boundingBox = [
                main_dict["horizontal_extent_lon_min"],
                main_dict["horizontal_extent_lat_max"],
            ]
            main_dict["item_bbox"] = list(map(float, boundingBox))
            main_dict["item_footprint"] = geometry.Point(
                main_dict["item_bbox"][0],
                main_dict["item_bbox"][1],
            )

            if main_dict["collection_footprint_point"] is None:
                main_dict["collection_footprint_point"] = main_dict[
                    "item_footprint"
                ]
            main_dict["collection_footprint_point"] = geometry.shape(
                main_dict["item_footprint"]
            ).union(geometry.shape(main_dict["collection_footprint_point"]))
            main_dict["collection_bbox"] = list(
                main_dict["collection_footprint_point"].bounds
            )

        elif linestring is True:
            boundingBox = [
                main_dict["horizontal_extent_lon_min"],
                main_dict["horizontal_extent_lat_min"],
                main_dict["horizontal_extent_lon_max"],
                main_dict["horizontal_extent_lat_max"],
            ]
            main_dict["item_bbox"] = list(map(float, boundingBox))
            main_dict["item_footprint"] = geometry.LineString(
                [
                    [main_dict["item_bbox"][0], main_dict["item_bbox"][1]],
                    [main_dict["item_bbox"][2], main_dict["item_bbox"][3]],
                ]
            )
            if main_dict["collection_footprint"] is None:
                main_dict["collection_footprint"] = main_dict["item_footprint"]
            main_dict["collection_footprint"] = geometry.shape(
                main_dict["item_footprint"]
            ).union(geometry.shape(main_dict["collection_footprint"]))
            main_dict["collection_bbox"] = list(
                main_dict["collection_footprint"].bounds
            )
        else:
            # TODO: we have to add another feature for LineString and Polygon
            boundingBox = [
                main_dict["horizontal_extent_lon_min"],
                main_dict["horizontal_extent_lat_min"],
                main_dict["horizontal_extent_lon_max"],
                main_dict["horizontal_extent_lat_max"],
            ]
            main_dict["item_bbox"] = list(map(float, boundingBox))
            main_dict["item_footprint"] = geometry.Polygon(
                [
                    [main_dict["item_bbox"][0], main_dict["item_bbox"][1]],
                    [main_dict["item_bbox"][0], main_dict["item_bbox"][3]],
                    [main_dict["item_bbox"][2], main_dict["item_bbox"][3]],
                    [main_dict["item_bbox"][2], main_dict["item_bbox"][1]],
                ]
            )
            if main_dict["collection_footprint"] is None:
                main_dict["collection_footprint"] = main_dict["item_footprint"]
            main_dict["collection_footprint"] = geometry.shape(
                main_dict["item_footprint"]
            ).union(geometry.shape(main_dict["collection_footprint"]))
            main_dict["collection_bbox"] = list(
                main_dict["collection_footprint"].bounds
            )



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

import logging
from typing import Any, Union

from .analysers.properties_verifier import Verifier


class Logger:
    """
    A class-based logger for TDS2STAC. It supports
    all the handlers from the standard python logging
    library.

    Args:
        logger_properties (dict, optional): Logger properties. Defaults to `dict()`.
            It's optional and has the following keys:

                logger_msg (str, optional)

                logger_handler (str, optional)

                logger_name (str, optional)

                logger_id (str, optional)

                logger_level (str, optional)

                logger_formatter (str, optional)

                logger_handler_host (str, optional)

                logger_handler_port (str, optional)

                logger_handler_url (str, optional)

                logger_handler_method (str, optional)

                logger_handler_secure (bool, optional)

                logger_handler_credentials (tuple, optional)

                logger_handler_context (tuple, optional)

                logger_handler_filename (str, optional)

                logger_handler_mode (str, optional)

                logger_handler_encoding (str, optional)

                logger_handler_delay (bool, optional)

                logger_handler_errors (str, optional)

                logger_handler_mailhost (str, optional)

                logger_handler_fromaddr (str, optional)

                logger_handler_toaddrs (str, optional)

                logger_handler_subject (str, optional)

                logger_handler_timeout (str, optional)
    """

    logger_properties: Union[dict[str, Any], None]
    """
    A dictionary that contains all the logger properties.

        It is optional and it is set to `None` by default.
        The following keys are supported:
                **logger_msg (str, optional):**
                    Logger message. Defaults to `None`.
                    But it is required when you want to log a message.
                **logger_handler (str, optional):**
                    Logger handler. Defaults to `NullHandler`.
                    Check the following website for more information:
                    https://docs.python.org/3/library/logging.handlers.html#module-logging.handlers
                **logger_name (str, optional):**
                    Logger name. Defaults to `TDS2STAC`. It's required
                    when you choose `HTTPHandler` as logger_handler.
                **logger_id (str, optional):**
                    Logger id. Defaults to `1`. It's required when you
                    choose `HTTPHandler` as logger_handler.
                **logger_level (str, optional):**
                    Logger level. Defaults to `DEBUG`. It's optional.
                    For more information check the following website:
                    https://docs.python.org/3/library/logging.html#levels
                **logger_formatter (str, optional):**
                    Logger format. Defaults to `%(levelname)-8s %(asctime)s \\t
                    %(filename)s @function %(funcName)s line %(lineno)s - %(message)s`.
                    For more information check the following website:
                    https://docs.python.org/3/library/logging.html#formatter-objects
                **logger_handler_host (str, optional):**
                    Logger host. Sets the value to 'None' by default.
                    It is required when `HTTPHandler` or `SocketHandler` are selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `HTTPHandler` or `SocketHandler` is
                    selected as the `logger_handler` value and neither `logger_handler_host` nor
                    `logger_handler_port` nor are specified.
                **logger_handler_port (str, optional):**
                    Logger port. Sets the value to 'None' by default.
                    It is required when `HTTPHandler` or `SocketHandler` are selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `HTTPHandler` or `SocketHandler` is
                    selected as the `logger_handler` value and neither `logger_handler_host` nor
                    `logger_handler_port` are specified.
                **logger_handler_url (str, optional):**
                    Logger url. Sets the value to 'None' by default.
                    It is required when `HTTPHandler` is selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `HTTPHandler` is
                    selected as the `logger_handler` value and neither `logger_handler_url`
                    is specified.
                **logger_handler_method (str, optional):**
                    HTTP methods. It supports sending logging messages
                    to a web server, using either GET or POST semantics. Sets the value to 'None' by default.
                    It is required when `HTTPHandler` is selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `HTTPHandler` is
                    selected as the `logger_handler` value and `logger_handler_method`
                    is not specified.
                **logger_handler_secure (bool, optional):**
                    HTTP secure. Sets the value to 'False' by default.
                    It is utilized when `HTTPHandler` or `SMTPHandler` are selected as the `logger_handler`.
                    But it is optional in both logger handlers.
                **logger_handler_credentials (tuple, optional):**
                    HTTP credentials. Sets the value to 'None' by default.
                    It is utilized when `HTTPHandler` or `SMTPHandler` are selected as the `logger_handler`.
                    But it is optional in both logger handlers.
                **logger_handler_context (tuple, optional):**
                    HTTP context. Sets the value to 'None' by default.
                    It is utilized when `HTTPHandler` is selected as the `logger_handler`.
                    But it is optional in both logger handlers.
                **logger_handler_filename (str, optional):**
                    File name. Sets the value to 'None' by default.
                    It is required when `FileHandler` or `WatchedFileHandler` are selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `FileHandler` or `WatchedFileHandler` is
                    selected as the `logger_handler` value and `logger_handler_filename` is not specified.
                **logger_handler_mode (str, optional):**
                    File mode. Sets the value to 'None' by default.
                    It is required when `FileHandler` or `WatchedFileHandler` are selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `FileHandler` or `WatchedFileHandler` is
                    selected as the `logger_handler` value and `logger_handler_mode` is not specified.
                **logger_handler_encoding (str, optional):**
                    File encoding. Sets the value to 'None' by default.
                    It is utilized when `FileHandler` or `WatchedFileHandler` are selected as the `logger_handler`.
                    But it is optional in both logger handlers.
                **logger_handler_delay (bool, optional):**
                    File delay. Sets the value to 'False' by default.
                    It is utilized when `FileHandler` or `WatchedFileHandler` are selected as the `logger_handler`.
                    But it is optional in both logger handlers.
                **logger_handler_errors (str, optional):**
                    File errors. Sets the value to 'None' by default.
                    It is utilized when `FileHandler` or `WatchedFileHandler` are selected as the `logger_handler`.
                    But it is optional in both logger handlers.
                **logger_handler_mailhost (str, optional):**
                    Mail host. Sets the value to 'None' by default.
                    It is required when `SMTPHandler` is selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `SMTPHandler` is
                    selected as the `logger_handler` value and `logger_handler_mailhost` is not specified.
                **logger_handler_fromaddr (str, optional):**
                    Mail from address. Sets the value to 'None' by default.
                    It is required when `SMTPHandler` is selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `SMTPHandler` is
                    selected as the `logger_handler` value and `logger_handler_fromaddr` is not specified.
                **logger_handler_toaddrs (str, optional):**
                    Mail to address. Sets the value to 'None' by default.
                    It is required when `SMTPHandler` is selected as the `logger_handler`.
                    The `logger_handler` will be set to 'NullHandler' if `SMTPHandler` is
                    selected as the `logger_handler` value and `logger_handler_toaddrs` is not specified.
                **logger_handler_subject (str, optional):**
                    Mail subject. Sets the value to 'None' by default.
                    It is utilized when `SMTPHandler` is selected as the `logger_handler`.
                    But it is optional in both logger handlers.
                **logger_handler_timeout (str, optional):**
                    Mail timeout. Sets the value to 'None' by default.
                    It is utilized when `SMTPHandler` is selected as the `logger_handler`.
                    But it is optional in both logger handlers.
    """

    def __init__(
        self,
        logger_properties: Union[dict[str, Any], None] = dict(),
    ):
        ##############################################
        # Setting the logger properties
        ##############################################

        verifier = Verifier()

        if logger_properties is not None and isinstance(
            logger_properties, dict
        ):
            verifier.logger_properties(logger_properties)
            ###############################################
            # Choosing the logger name and ID if they are
            # not set when the HTTPHandler is selected as
            # the logger handler
            ###############################################
            if logger_properties.get("logger_name") is None:
                logger_properties["logger_name"] = "TDS2STAC"
            if logger_properties.get("logger_id") is None:
                logger_properties["logger_id"] = "1"
            ###############################################
            # Specifying the standard logger parameters for
            # all handlers
            ###############################################
            self.logger = logging.getLogger(
                str(logger_properties["logger_name"])
                + "_"
                + str(logger_properties["logger_id"])
            )
            if logger_properties.get("logger_level") is None:
                self.logger.setLevel(level=logging.DEBUG)
                logger_properties["logger_level"] = "DEBUG"
            elif logger_properties.get(
                "logger_level"
            ) is not None and logger_properties.get("logger_level") in [
                "NOTSET",
                "DEBUG",
                "INFO",
                "WARNING",
                "ERROR",
                "CRITICAL",
            ]:
                self.logger.setLevel(level=logger_properties["logger_level"])
            if logger_properties.get("logger_formatter") is None:
                logFormatter = logging.Formatter(
                    fmt="%(levelname)-8s %(asctime)s \t %(filename)s @function %(funcName)s line %(lineno)s - %(message)s",
                )
            elif logger_properties.get("logger_formatter") is not None:
                logFormatter = logging.Formatter(
                    fmt=logger_properties["logger_formatter"],
                )
            ###############################################
            # Verifying that the logger handler is supported
            # by the current version of TDS2STAC
            ###############################################
            if logger_properties.get(
                "logger_handler"
            ) is not None and logger_properties.get("logger_handler") not in [
                "NullHandler",
                "StreamHandler",
                "HTTPHandler",
                "FileHandler",
                "WatchedFileHandler",
                "SocketHandler",
                "SMTPHandler",
            ]:
                print(
                    "Warning: The current version of TDS2STAC is incompatible with your Logging \
                    Handler. It will be set automatically to `NullHandler`"
                )
                self.Null_Handler()
                return
            ###############################################
            # Self.legger is defining here depending on the
            # `logger_handler`
            ###############################################
            if (
                logger_properties.get("logger_handler") is None
                or logger_properties.get("logger_handler") == "NullHandler"
            ):
                self.Null_Handler()
            elif logger_properties.get("logger_handler") == "StreamHandler":
                streamHandler = logging.StreamHandler()
                streamHandler.setFormatter(logFormatter)
                streamHandler.setLevel(logging.DEBUG)
                if self.logger.handlers == [] or not any(
                    "StreamHandler"
                    in self.logger.handlers[i].__class__.__name__
                    for i in range(len(self.logger.handlers))
                ):
                    self.logger.addHandler(streamHandler)
            elif logger_properties.get("logger_handler") == "SocketHandler":
                if (
                    logger_properties.get("logger_handler_host") is None
                    or logger_properties.get("logger_handler_port") is None
                ):
                    print(
                        "Warning: You must specify `logger_handler_host` and `logger_handler_port` \
                        when selecting `SocketHandler` as the `logger_handler`. if not, `logger_handler` \
                        will transition to `NullHandler`"
                    )
                    self.Null_Handler()
                    return
                socketHandler = logging.handlers.SocketHandler(  # type: ignore
                    host=logger_properties.get("logger_handler_host"),
                    port=logger_properties.get("logger_handler_port"),
                )
                socketHandler.setFormatter(logFormatter)
                socketHandler.setLevel(logging.DEBUG)
                if self.logger.handlers == [] or not any(
                    "SocketHandler"
                    in self.logger.handlers[i].__class__.__name__
                    for i in range(len(self.logger.handlers))
                ):
                    self.logger.addHandler(socketHandler)
            elif logger_properties.get("logger_handler") == "HTTPHandler":
                if (
                    logger_properties.get("logger_handler_host") is None
                    or logger_properties.get("logger_handler_port") is None
                    or logger_properties.get("logger_handler_url") is None
                    or logger_properties.get("logger_handler_method") is None
                ):
                    print(
                        "Warning: You must specify `logger_handler_host`, `logger_handler_port`, \
                        `logger_handler_url`, `logger_handler_method` when selecting `HTTPHandler` \
                        as the `logger_handler`. if not, `logger_handler` will transition to `NullHandler`"
                    )
                    self.Null_Handler()
                    return
                httpHandler = logging.handlers.HTTPHandler(  # type: ignore
                    host=str(logger_properties.get("logger_handler_host"))
                    + ":"
                    + str(logger_properties.get("logger_handler_port")),
                    url=logger_properties.get("logger_handler_url"),
                    method=logger_properties.get("logger_handler_method"),
                    secure=logger_properties.get("logger_handler_secure"),
                    credentials=logger_properties.get(
                        "logger_handler_credentials"
                    ),
                    context=logger_properties.get("logger_handler_context"),
                )
                httpHandler.setFormatter(logFormatter)
                httpHandler.setLevel(logging.DEBUG)
                if self.logger.handlers == [] or not any(
                    "HTTPHandler" in self.logger.handlers[i].__class__.__name__
                    for i in range(len(self.logger.handlers))
                ):
                    self.logger.addHandler(httpHandler)
            elif logger_properties.get("logger_handler") == "FileHandler":
                if (
                    logger_properties.get("logger_handler_filename") is None
                    or logger_properties.get("logger_handler_mode") is None
                ):
                    print(
                        "Warning: You must specify `logger_handler_filename` and `logger_handler_mode` \
                        when selecting `FileHandler` as the `logger_handler`. if not, `logger_handler`\
                          will transition to `NullHandler`"
                    )
                    self.Null_Handler()
                    return
                fileHandler = logging.FileHandler(
                    filename=str(
                        logger_properties.get("logger_handler_filename")
                    ),
                    mode=str(logger_properties.get("logger_handler_mode")),
                    encoding=logger_properties.get("logger_handler_encoding"),
                    delay=bool(logger_properties.get("logger_handler_delay")),
                    errors=logger_properties.get("logger_handler_errors"),
                )
                fileHandler.setFormatter(logFormatter)
                fileHandler.setLevel(logging.DEBUG)
                if self.logger.handlers == [] or not any(
                    "FileHandler" in self.logger.handlers[i].__class__.__name__
                    for i in range(len(self.logger.handlers))
                ):
                    self.logger.addHandler(fileHandler)
            elif logger_properties.get("logger_handler") == "SMTPHandler":
                if (
                    logger_properties.get("logger_handler_mailhost") is None
                    or logger_properties.get("logger_handler_fromaddr") is None
                    or logger_properties.get("logger_handler_toaddrs") is None
                    or logger_properties.get("logger_handler_subject") is None
                ):
                    print(
                        "Warning: You must specify `logger_handler_mailhost`, `logger_handler_fromaddr`, \
                        `logger_handler_toaddrs`, `logger_handler_subject` when selecting `SMTPHandler` \
                        as the `logger_handler`. if not, `logger_handler` will transition to `NullHandler`"
                    )
                    self.Null_Handler()
                    return
                smtpHandler = logging.handlers.SMTPHandler(  # type: ignore
                    mailhost=logger_properties.get("logger_handler_mailhost"),
                    fromaddr=logger_properties.get("logger_handler_fromaddr"),
                    toaddrs=logger_properties.get("logger_handler_toaddrs"),
                    subject=logger_properties.get("logger_handler_subject"),
                    credentials=logger_properties.get(
                        "logger_handler_credentials"
                    ),
                    secure=logger_properties.get("logger_handler_secure"),
                    timeout=logger_properties.get("logger_handler_timeout"),
                )
                smtpHandler.setFormatter(logFormatter)
                smtpHandler.setLevel(logging.DEBUG)
                if self.logger.handlers == [] or not any(
                    "SMTPHandler" in self.logger.handlers[i].__class__.__name__
                    for i in range(len(self.logger.handlers))
                ):
                    self.logger.addHandler(smtpHandler)

            if logger_properties.get("logger_msg") is not None:
                if logger_properties.get("logger_level") is not None and (
                    logger_properties.get("logger_level") == "DEBUG"
                    or logger_properties.get("logger_level") == "NOTSET"
                ):
                    self.logger.debug(logger_properties.get("logger_msg"))
                elif (
                    logger_properties.get("logger_level") is not None
                    and logger_properties.get("logger_level") == "INFO"
                ):
                    self.logger.info(logger_properties.get("logger_msg"))
                elif (
                    logger_properties.get("logger_level") is not None
                    and logger_properties.get("logger_level") == "WARNING"
                ):
                    self.logger.warning(logger_properties.get("logger_msg"))
                elif (
                    logger_properties.get("logger_level") is not None
                    and logger_properties.get("logger_level") == "ERROR"
                ):
                    self.logger.error(logger_properties.get("logger_msg"))
                elif (
                    logger_properties.get("logger_level") is not None
                    and logger_properties.get("logger_level") == "CRITICAL"
                ):
                    self.logger.critical(logger_properties.get("logger_msg"))

    def Null_Handler(self):
        """
        This is a function to return a NullHandler
        """
        nullhandler = logging.NullHandler()
        self.logger.handlers = []
        self.logger.addHandler(nullhandler)



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

"""Setup script for the tds2stac package."""
import versioneer
from setuptools import setup

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
import sys
import traceback
from datetime import datetime
from typing import Tuple, Union
from urllib.parse import quote_plus
from urllib.request import urlopen

import pytz
import requests
from dateutil.parser import parse
from lxml import etree

from . import logger, utils
from .analysers.recognizer import Recognizer
from .dimensions.spatial import Spatial
from .dimensions.temporal import Temporal
from .statics import constants
from .webservices import core


class CollectionHarvester(object):
    """
    This class harvests data pertaining to Collections
    from TDS catalogs. Depending on the sort of dataset
    scenario, it returns one of the five variables below.
    `collection_id` , `collection_title` , `collection_description`
    , `collection_url` , and `collection_subdirs`.

    Args:
        url (str): TDS catalog URL address
        recognizer (str): `status` senario number of :class:`~tds2stac.Recognizer`
        subdirs (list): subdirs is a list of url, id, title, and subdirs of a nested dataset
        collection_tuples (list):  a tuple of STAC collection's auto-generated ID, user-ID, user-Title and user-Description defined by user.
        auth (tuple): Authentication tuple for TDS catalog e.g.('user', 'password') (optional)
        logger_properties (dict): dictionary of logger properties


    """

    url: str
    """
        TDS catalog URL address. Initial point of harvesting e.g.
        https://thredds.atmohub.kit.edu/thredds/catalog/caribic/IAGOS-CARIBIC_MS_files_collection_20231017/catalog.html (*)
    """
    recognizer: Union[str, None]
    """
        `status` senario output of Recognizer class
    """
    subdirs: Union[list, None]
    """
        subdirs is a list of url, id, title, and subdirs of a nested dataset
    """
    collection_tuples: Union[list[tuple], None]
    """
        a tuple of STAC collection's auto-generated ID, user-ID, user-Title and user-Description defined by user.
    """
    auth: Union[Tuple[str, str], None]
    """
        Authentication tuple for TDS catalog e.g.('user', 'password') (optional)
    """
    logger_properties: dict
    """
        dictionary of logger properties, more information in :class:`~tds2stac.logger.Logger`
    """

    def __init__(
        self,
        url: str,
        recognizer: Union[str, None],
        subdirs: Union[list, None] = [],
        collection_tuples: Union[list[tuple], None] = None,
        auth: Union[Tuple[str, str], None] = None,
        logger_properties: dict = dict(),
    ):
        ###############################################
        # Defining output and input variables as an empty parameter
        ###############################################
        self.auth = auth
        self.collection_tuples = collection_tuples
        self.Recognizer_output = recognizer
        self.collection_id: Union[str, None] = None
        self.collection_title: Union[str, None] = None
        self.colleciton_description: Union[str, None] = None
        self.collection_url: Union[str, None] = None
        self.collection_subdirs: Union[list, None] = []
        if logger_properties is not None:
            self.logger_properties = logger_properties

        ###############################################
        # Getting the Collection ID and Descriptionin
        # in none nested scenarios
        ###############################################
        if any(
            scenario in str(self.Recognizer_output)
            for scenario in constants.none_nested_scenarios
        ):
            (
                self.collection_id,
                self.collection_title,
                self.collection_description,
            ) = self.collection_id_desc_maker(
                url, self.collection_tuples, self.Recognizer_output
            )
            self.collection_url = url
        ###############################################
        # Getting the Collection ID and Descriptionin
        # in nested scenarios
        ###############################################
        if any(
            scenario in str(self.Recognizer_output)
            for scenario in constants.nested_scenarios
        ):
            #####################################
            # Collection cases with subdirs
            #####################################
            if subdirs is not None:
                if subdirs[3] != []:
                    self.collection_url = subdirs[0]
                    self.collection_id = subdirs[1]
                    self.collection_title = subdirs[2]
                    self.collection_subdirs = subdirs[3]
                    self.collection_description = (
                        "[Link to TDS](" + url.replace("xml", "html") + ")"
                    )
                    if self.collection_tuples is not None:
                        for i in self.collection_tuples:
                            if i[0] == self.collection_id:
                                if i[1] != "" and not i[1].isspace():
                                    self.collection_id = i[1]
                                else:
                                    self.collection_id = self.collection_id
                                if i[2] != "" and not i[2].isspace():
                                    self.collection_title = i[2]
                                else:
                                    self.collection_title = (
                                        self.collection_title
                                    )
                                if i[3] != "" and not i[3].isspace():
                                    self.collection_description = (
                                        i[3]
                                        .strip()
                                        .replace("\n", "\n\n")
                                        .replace('"', "")
                                        + "\n\n [Link to TDS]("
                                        + url.replace("xml", "html")
                                        + ")"
                                    )
                #####################################
                # Collection cases without subdirs
                #####################################
                else:
                    xml_url_catalog, id_catalog, xml = utils.xml_processing(
                        subdirs[0], self.auth
                    )
                    recognizer_output_temp = Recognizer(subdirs[0])
                    self.collection_url = subdirs[0]
                    (
                        self.collection_id,
                        self.collection_title,
                        self.collection_description,
                    ) = self.collection_id_desc_maker(
                        xml_url_catalog,
                        self.collection_tuples,
                        recognizer_output_temp.status,
                    )

    def collection_id_desc_maker(
        self,
        url: str,
        collection_tuples: Union[list[tuple], None] = None,
        recognizer_output: Union[str, None] = None,
    ):
        """
        A function for getting collection id and description
        from the TDS catalog urls and pre-defined `collection_tuples`
        for scenarios number 4, 5, 6, 7 ,and 9

        Args:
            url (str): TDS catalog URL address
            collection_tuples (list):  a tuple of STAC collection's auto-generated ID, user-ID, user-Title and user-Description defined by user.
            recognizer_output (str): `status` senario output of Recognizer class
        """

        if "Fourth Scenario" in str(recognizer_output):
            collection_id = utils.replacement_func_collection_item_id(
                url + "-empty".replace("html", "xml")
            )
            collection_title = utils.replacement_func(
                url + " Empty".replace("html", "xml")
            )
            collection_description = (
                "[Link to TDS](" + url.replace("xml", "html") + ")"
            )
            if collection_tuples is not None:
                for i in collection_tuples:
                    if i[0] == collection_id:
                        if i[1] != "" and not i[1].isspace():
                            collection_id = i[1]
                        else:
                            collection_id = collection_id
                        if i[2] != "" and not i[2].isspace():
                            collection_title = i[2]
                        else:
                            collection_title = collection_title
                        if i[3] != "" and not i[3].isspace():
                            collection_description = (
                                i[3]
                                .strip()
                                .replace("\n", "\n\n")
                                .replace('"', "")
                                + "\n\n [Link to TDS]("
                                + url.replace("xml", "html")
                                + ")"
                            )
            return collection_id, collection_title, collection_description
        fifth_and_ninth_scenarios = ["Fifth Scenario", "Ninth Scenario"]
        if any(x in str(recognizer_output) for x in fifth_and_ninth_scenarios):
            collection_id = utils.replacement_func_collection_item_id(
                url + "-collection".replace("html", "xml")
            )
            collection_title = utils.replacement_func(
                url + " Collection".replace("html", "xml")
            )
            collection_description = (
                "[Link to TDS](" + url.replace("xml", "html") + ")"
            )
            if collection_tuples is not None:
                for i in collection_tuples:
                    if i[0] == collection_id:
                        if i[1] != "" and not i[1].isspace():
                            collection_id = i[1]
                        else:
                            collection_id = collection_id
                        if i[2] != "" and not i[2].isspace():
                            collection_title = i[2]
                        else:
                            collection_title = collection_title
                        if i[3] != "" and not i[3].isspace():
                            collection_description = (
                                i[3]
                                .strip()
                                .replace("\n", "\n\n")
                                .replace('"', "")
                                + "\n\n [Link to TDS]("
                                + url.replace("xml", "html")
                                + ")"
                            )
            return collection_id, collection_title, collection_description

        if "Sixth Scenario" in str(recognizer_output):
            collection_id = utils.replacement_func_collection_item_id(
                url + "-single".replace("html", "xml")
            )
            collection_title = utils.replacement_func(
                url + " Single".replace("html", "xml")
            )
            collection_description = (
                "[Link to TDS](" + url.replace("xml", "html") + ")"
            )
            if collection_tuples is not None:
                for i in collection_tuples:
                    if i[0] == collection_id:
                        if i[1] != "" and not i[1].isspace():
                            collection_id = i[1]
                        else:
                            collection_id = collection_id
                        if i[2] != "" and not i[2].isspace():
                            collection_title = i[2]
                        else:
                            collection_title = collection_title
                        if i[3] != "" and not i[3].isspace():
                            collection_description = (
                                i[3]
                                .strip()
                                .replace("\n", "\n\n")
                                .replace('"', "")
                                + "\n\n [Link to TDS]("
                                + url.replace("xml", "html")
                                + ")"
                            )

            return collection_id, collection_title, collection_description

        if "Seventh Scenario" in str(recognizer_output):
            collection_id = utils.replacement_func_collection_item_id(
                url + "-aggregated".replace("html", "xml")
            )
            collection_title = utils.replacement_func(
                url + " Aggregated".replace("html", "xml")
            )
            collection_description = (
                "[Link to TDS](" + url.replace("xml", "html") + ")"
            )
            if collection_tuples is not None:
                for i in collection_tuples:
                    if i[0] == collection_id:
                        if i[1] != "" and not i[1].isspace():
                            collection_id = i[1]
                        else:
                            collection_id = collection_id
                        if i[2] != "" and not i[2].isspace():
                            collection_title = i[2]
                        else:
                            collection_title = collection_title
                        if i[3] != "" and not i[3].isspace():
                            collection_description = (
                                i[3]
                                .strip()
                                .replace("\n", "\n\n")
                                .replace('"', "")
                                + "\n\n [Link to TDS]("
                                + url.replace("xml", "html")
                                + ")"
                            )

            return collection_id, collection_title, collection_description

    def __iter__(self):
        yield "collection_id", self.collection_id
        yield "collection_title", self.collection_title
        yield "collection_description", self.collection_description
        yield "collection_url", self.collection_url
        yield "collection_subdirs", self.collection_subdirs


class ItemHarvester(object):
    """
    This class harvests information about an Item
    from TDS data catalogs. It ultimately returns
    a dictionary of harvesting variables, based on
    the type of dataset scenario and activated
    extensions.

    Args:
        url (str): TDS catalog URL address
        elem (str): xml element of the data in dataset
        harvesting_vars (dict): dictionary of harvesting variables that is going to be filled
        web_service_dict (dict): web service that the user wants to harvest from
        datetime_after (str): datetime that the user wants to harvest data after that
        datetime_before (str): datetime that the user wants to harvest data before that
        spatial_information (list): Spatial information of 2D datasets e.g. [minx, maxx, miny, maxy] or 1D dataset e.g. [x,y]
        temporal_format_by_dataname (str): datetime format for datasets that have datetime in their name e.g `e%y%m%d%H.%M%S%f`(optional),
        extension_properties (dict): dictionary of extension properties (optional)
        linestring (bool): using this attribute, user activate making LineString instead of Polygon (True and False)  (optional)
        logger_properties (dict): dictionary of logger properties
    """

    usl: str
    """
        TDS catalog URL address. Initial point of harvesting e.g.
        https://thredds.atmohub.kit.edu/thredds/catalog/caribic/IAGOS-CARIBIC_MS_files_collection_20231017/catalog.html (*)
    """
    elem: etree.Element
    """
        xml element of the data in dataset. It's an element of the
        xml file that is going to be harvested
    """
    harvesting_vars: dict
    """
        dictionary of harvesting variables that is going to be filled
    """
    web_service_dict: Union[dict, None]
    """
        web service that the user wants to harvest from
    """
    datetime_after: Union[str, None]
    """
        datetime that the user wants to harvest data after that
    """
    datetime_before: Union[str, None]
    """
        datetime that the user wants to harvest data before that
    """
    spatial_information: Union[list, None]
    """
        Spatial information of 2D datasets e.g. [minx, maxx, miny, maxy] or 1D dataset e.g. [x,y] (optional)
    """
    temporal_format_by_dataname: Union[str, None]
    """
        datetime format for datasets that have datetime in their name e.g `e%y%m%d%H.%M%S%f`(optional)
    """
    extension_properties: Union[dict, None]
    """
        dictionary of extension properties (optional)
    """
    linestring: bool
    """
        using this attribute, user activate making LineString instead of Polygon (True and False)  (optional)
    """
    logger_properties: dict
    """
        dictionary of logger properties, more information in :class:`~tds2stac.logger.Logger`
    """

    def __init__(
        self,
        url: str,
        elem: etree.Element,
        harvesting_vars: dict,
        web_service_dict: Union[dict, None],
        datetime_after: Union[datetime, None] = None,
        datetime_before: Union[datetime, None] = None,
        spatial_information: Union[list, None] = None,
        temporal_format_by_dataname: Union[str, None] = None,
        extension_properties: Union[dict, None] = None,
        linestring: bool = False,
        logger_properties: dict = dict(),
    ):
        # Defining variables in TDS catalog and webservices as an empty parameter
        # Getting the item_id
        self.harvesting_vars = harvesting_vars
        self.harvesting_vars["item_id"] = utils.replacement_func(
            elem.get("ID")
        )
        if logger_properties is not None:
            self.logger_properties = logger_properties

        #################################
        # Opening the xml file
        #################################
        if "?dataset=" in url:
            data_get = requests.get(
                str(url),
                None,
                verify=False,
            )
        else:
            data_get = requests.get(
                str(url) + "?dataset=" + str(elem.get("ID")),
                None,
                verify=False,
            )
        #################################
        # Getting the tree of the xml file
        #################################
        try:
            tree_data = etree.XML(data_get.text.encode("utf-8"))
        except etree.XMLSyntaxError:
            ex_type, ex_value, ex_traceback = sys.exc_info()
            if ex_type is not None and ex_value is not None:
                self.logger_properties[
                    "logger_msg"
                ] = "`datetime_filter` warning: %s : %s" % (
                    ex_type.__name__,
                    ex_value,
                )
                self.logger_properties["logger_level"] = "WARNING"
            logger.Logger(self.logger_properties)
            return
        else:
            try:
                ###############################################
                # It serves as a function to skip
                # data based on `datetime_before` and `datetime_after`
                ###############################################
                extracted_date = elem.find(
                    './/{%s}date[@type="modified"]' % constants.unidata
                )
                if extracted_date is not None:
                    try:
                        extracted_date = extracted_date.text
                        dt = parse(extracted_date)
                        self.harvesting_vars[
                            "modified_date_time"
                        ] = dt.replace(tzinfo=pytz.utc)
                    except ValueError:
                        ex_type, ex_value, ex_traceback = sys.exc_info()
                        if ex_type is not None and ex_value is not None:
                            self.logger_properties[
                                "logger_msg"
                            ] = "`datetime_filter` error: %s : %s" % (
                                ex_type.__name__,
                                ex_value,
                            )
                            self.logger_properties["logger_level"] = "ERROR"
                            logger.Logger(self.logger_properties)
                        return
                    else:
                        dt = dt.replace(tzinfo=pytz.utc)
                        if datetime_after and dt < datetime_after:
                            return
                        if datetime_before and dt > datetime_before:
                            return
                # bleedfixing datatime in STAC-Colleciton and STAC-Items for aggregated datasets. Should be fixed in the future
                else:
                    self.harvesting_vars["modified_date_time"] = parse(
                        datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fz")
                    ).replace(tzinfo=pytz.utc)
                    self.logger_properties[
                        "logger_msg"
                    ] = "The dataset does not have a modified datetime attribute so it is set to the current datetime"
                    self.logger_properties["logger_level"] = "INFO"
                    logger.Logger(self.logger_properties)
                ###############################################
                # Getting `dataset`, `catalog_url` and `catalog_id` of the data
                ###############################################
                self.harvesting_vars["dataset"] = tree_data.find(
                    "{%s}dataset" % constants.unidata
                )
                self.harvesting_vars["catalog_url"] = url.split("?")[0]
                self.harvesting_vars["main_dataset_url"] = url
                harvesting_vars["catalog_id"] = self.harvesting_vars[
                    "dataset"
                ].get("ID")
                metadata = self.harvesting_vars["dataset"].find(
                    "{%s}metadata" % constants.unidata
                )
                ########################################
                # Finding sevice tag in data tree element
                ########################################
                service_tag = self.harvesting_vars["dataset"].find(
                    "{%s}serviceName" % constants.unidata
                )
                if service_tag is None and metadata is not None:
                    service_tag = metadata.find(
                        "{%s}serviceName" % constants.unidata
                    )
                if service_tag is None:
                    # Use services found in the file depends on the version of TDS 4 or 5
                    try:
                        self.harvesting_vars["services"] = tree_data.findall(
                            ".//{%s}service[@serviceType='Compound']"
                            % constants.unidata
                        )
                    except Exception:
                        self.harvesting_vars["services"] = tree_data.findall(
                            ".//{%s}service[@serviceType='compound']"
                            % constants.unidata
                        )
                else:
                    # Use specific named services
                    self.harvesting_vars["services"] = tree_data.findall(
                        ".//{%s}service[@name='%s']"
                        % (constants.unidata, service_tag.text)
                    )
                #############################################
                # Loop over all existing services in the data
                #############################################
                if "?dataset=" in url:
                    harvesting_vars["service_url_html"] = utils.xml2html(
                        str(url)
                    )
                else:
                    harvesting_vars["service_url_html"] = utils.xml2html(
                        str(url) + "?dataset=" + str(elem.get("ID"))
                    )
                if web_service_dict is not None:
                    if (
                        web_service_dict.get("web_service_config_file")
                        == "default"
                        or web_service_dict.get("web_service_config_file")
                        is None
                    ):
                        json_file = constants.default_tag_config_file
                    else:
                        json_file = web_service_dict["web_service_config_file"]
                extension_list = ["main_metadata"]
                # TODO: we need to think sabout this part to make it more specific for an extension
                if extension_properties is not None and isinstance(
                    extension_properties.get("item_extensions"), list
                ):
                    for item_extension in extension_properties[
                        "item_extensions"
                    ]:
                        if isinstance(item_extension, str):
                            extension_list.append(item_extension)
                        elif isinstance(item_extension, tuple):
                            extension_list.append(item_extension[0])
                if extension_properties is not None and isinstance(
                    extension_properties.get("collection_extensions"), list
                ):
                    for collection_extension in extension_properties[
                        "collection_extensions"
                    ]:
                        if isinstance(collection_extension, str):
                            extension_list.append(collection_extension)
                        elif isinstance(collection_extension, tuple):
                            extension_list.append(collection_extension[0])

                # if extension_properties is not None and extension_properties["item_common_metadata"] is True:
                #     extension_list.extend("common_metadata")
                # if extension_properties is not None and (extension_properties["item_scientific"] is True or extension_properties["collection_scientific"] is True):
                #     extension_list.extend("scientific_extension")
                for i, service in enumerate(self.harvesting_vars["services"]):
                    if (
                        service.get("serviceType") == "Compound"
                        or service.get("serviceType") == "compound"
                    ):
                        services_list = service.findall(
                            "{%s}service" % constants.unidata
                        )
                        services_list.append(url)
                        for s in services_list:
                            if s == url:
                                s_get_name = "catalog"
                                service_url = url
                            else:
                                service_url = utils.references_urls(
                                    url, s.get("base")
                                ) + self.harvesting_vars["dataset"].get(
                                    "urlPath"
                                )

                                s_get_name = s.get("name")
                                if s.get("suffix") is not None:
                                    service_url += s.get("suffix")
                                if s_get_name in ["iso", "ncml", "uddc"]:
                                    service_url += (
                                        "?dataset=%s&&catalog=%s"
                                        % (
                                            self.harvesting_vars["catalog_id"],
                                            quote_plus(
                                                self.harvesting_vars[
                                                    "catalog_url"
                                                ]
                                            ),
                                        )
                                    )
                                if s_get_name in ["wms"]:
                                    service_url += "?service=WMS&version=1.3.0&request=GetCapabilities"
                                if s_get_name in ["dap4"]:
                                    service_url += ".dmr.xml"

                            if web_service_dict is not None and (
                                web_service_dict.get("webservices") is not None
                                and all(
                                    a.lower()
                                    in constants.supported_webservices
                                    for a in web_service_dict["webservices"]
                                )
                                and any(
                                    "/thredds/" + b.lower() + "/"
                                    in service_url
                                    for b in web_service_dict["webservices"]
                                )
                            ):
                                # To prevent 400 HTTP error
                                try:
                                    root = etree.XML(
                                        urlopen(service_url).read()
                                    )
                                    # root_nested = etree.parse(urlopen(service_url)).getroot()
                                except Exception:
                                    (
                                        ex_type,
                                        ex_value,
                                        ex_traceback,
                                    ) = sys.exc_info()
                                    if (
                                        ex_type is not None
                                        and ex_value is not None
                                    ):
                                        self.logger_properties[
                                            "logger_msg"
                                        ] = (
                                            "400 HTTP verifier error code `urlopen`: %s: %s . It appears that the %s webservice you are utilizing in your configuration is deactivated or not working. Please check the `tag_config.json` file or ensure the web service is activated."
                                            % (
                                                ex_type.__name__,
                                                ex_value,
                                                s_get_name,
                                            )
                                        )
                                        self.logger_properties[
                                            "logger_level"
                                        ] = "ERROR"
                                        logger.Logger(self.logger_properties)
                                    print(traceback.format_exc())
                                    continue

                                core.WebServiceContentScraper(
                                    root,
                                    json_file,
                                    extension_list,
                                    harvesting_vars,
                                )
                            # elif web_service_dict is not None and web_service_dict["webservice"] == "all":
                            #     if s_get_name in constants.supported_webservices:
                            #         # To prevent 400 HTTP error
                            #         try:
                            #             root = etree.XML(urlopen(service_url).read())
                            #         except Exception:
                            #             (
                            #                 ex_type,
                            #                 ex_value,
                            #                 ex_traceback,
                            #             ) = sys.exc_info()
                            #             if ex_type is not None and ex_value is not None:
                            #                 self.logger_properties["logger_msg"] = "400 HTTP verifier error code `urlopen`: %s: %s . It appears that the %s webservice you are utilizing in your configuration is deactivated or not working. Please check the `tag_config.json` file or ensure the web service is activated." % (
                            #                     ex_type.__name__,
                            #                     ex_value,
                            #                     s_get_name,
                            #                 )
                            #                 self.logger_properties["logger_level"] = "ERROR"
                            #                 logger.Logger(self.logger_properties)
                            #             # print(traceback.format_exc())
                            #             continue
                            #         core.WebServiceContentScraper(
                            #             root,
                            #             json_file,
                            #             extension_list,
                            #             harvesting_vars,
                            #         )
                            # elif web_service_dict is not None and web_service_dict["webservice"] == "catalog":
                            #     try:
                            #         root = etree.XML(urlopen(url).read())
                            #     except Exception:
                            #         (
                            #             ex_type,
                            #             ex_value,
                            #             ex_traceback,
                            #         ) = sys.exc_info()
                            #         if ex_type is not None and ex_value is not None:
                            #             self.logger_properties["logger_msg"] = "400 HTTP verifier error code `urlopen`: %s: %s . It appears that the %s webservice you are utilizing in your configuration is deactivated or not working. Please check the `tag_config.json` file or ensure the web service is activated." % (
                            #                 ex_type.__name__,
                            #                 ex_value,
                            #                 s_get_name,
                            #             )
                            #             self.logger_properties["logger_level"] = "ERROR"
                            #             logger.Logger(self.logger_properties)
                            #         print(traceback.format_exc())
                            #         continue
                            #     core.WebServiceContentScraper(
                            #         root,
                            #         json_file,
                            #         extension_list,
                            #         harvesting_vars,
                            #     )
                            elif web_service_dict is not None and (
                                web_service_dict.get("webservices") is not None
                                and not all(
                                    a.lower()
                                    in constants.supported_webservices
                                    for a in web_service_dict["webservices"]
                                )
                                and not any(
                                    "/thredds/" + b.lower() + "/"
                                    in service_url
                                    for b in web_service_dict["webservices"]
                                )
                            ):
                                self.logger_properties[
                                    "logger_msg"
                                ] = "It couldn't find any activated web service according to your `tag_config.json` configuration. Please check your web service configuration file, `tag_config.json`"
                                self.logger_properties[
                                    "logger_level"
                                ] = "ERROR"
                                logger.Logger(self.logger_properties)
                                continue
                    spatial = Spatial()
                    spatial_output = spatial.regulator(
                        self.harvesting_vars, spatial_information
                    )
                    if isinstance(spatial_output, list):
                        spatial_output = list(dict.fromkeys(spatial_output))
                        for error in spatial_output:
                            self.logger_properties["logger_msg"] = error
                            self.logger_properties["logger_level"] = "ERROR"
                            logger.Logger(self.logger_properties)
                        # important break, because spatial infos is one of mandatory elements of STAC
                        continue
                    else:
                        (
                            self.harvesting_vars["horizontal_extent_lon_min"],
                            self.harvesting_vars["horizontal_extent_lon_max"],
                            self.harvesting_vars["horizontal_extent_lat_min"],
                            self.harvesting_vars["horizontal_extent_lat_max"],
                        ) = spatial_output

                        spatial.harvester(self.harvesting_vars, linestring)
                    #########################################
                    # Harvesting and then regulating temporal
                    # information from the data
                    #########################################
                    temporal_class = Temporal()
                    temporal_output = temporal_class.regulator(
                        self.harvesting_vars,
                        temporal_format_by_dataname,
                        elem.get("name"),
                    )
                    if isinstance(temporal_output, list):
                        temporal_output = list(dict.fromkeys(temporal_output))
                        for error in temporal_output:
                            self.logger_properties["logger_msg"] = error
                            self.logger_properties["logger_level"] = "ERROR"
                            logger.Logger(self.logger_properties)
                        # important break, because temporal infos is one of mandatory elements of STAC
                        continue
                    self.harvesting_vars["collection_interval_time_final"] = [
                        self.harvesting_vars["collection_interval_time"][0],
                        self.harvesting_vars["collection_interval_time"][-1],
                    ]

            except Exception:
                ex_type, ex_value, ex_traceback = sys.exc_info()
                if ex_type is not None and ex_value is not None:
                    self.logger_properties[
                        "logger_msg"
                    ] = "`ItemHarvester` error: %s : %s" % (
                        ex_type.__name__,
                        ex_value,
                    )
                    self.logger_properties["logger_level"] = "CRITICAL"
                    logger.Logger(self.logger_properties)
                print(traceback.format_exc())
                return

    def __iter__(self):
        for key in self.harvesting_vars.keys():
            yield key, self.harvesting_vars[key]
